""" ------------------------------
    SET OF FUNCTIONS TO VISUALIZE AND DOWNLOAD IMAGES FROM ALLEN MOUSE BRAIN DATASETS:
        - Download images from experiments (e.g. ISH gene expression images)
        - Download SVG/PNG version of the Allen Brain Atlas images (for any atlas)
"""
import sys

sys.path.append("./")

from pathlib import Path
from os import chdir
import pandas as pd
from rich.progress import track

try:
    from allensdk.api.queries.svg_api import SvgApi
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "You need to install allensdk for this to work: `pip install allensdk`"
    )
else:
    from allensdk.api.queries.image_download_api import ImageDownloadApi
    from allensdk.api.queries.annotated_section_data_sets_api import (
        AnnotatedSectionDataSetsApi,
    )
    from allensdk.api.queries.ontologies_api import OntologiesApi

    allensdk_installed = True

from brainrender.Utils.webqueries import send_query
from brainrender.Utils.decorators import fail_on_no_connection


class ImageDownload(SvgApi, ImageDownloadApi):
    """ 
    Handles query to the Allen ImageDownloadApi and saves the data
    """

    mouse_coronal = "Mouse, P56, Coronal"
    mouse_sagittal = "Mouse, P56, Sagittal"
    mouse3d = "Mouse, Adult, 3D Coronal"

    # useful tutorial: https://allensdk.readthedocs.io/en/latest/_static/examples/nb/image_download.html
    @fail_on_no_connection
    def __init__(self):
        SvgApi.__init__(
            self
        )  # https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/api/queries/svg_api.py
        ImageDownloadApi.__init__(
            self
        )  # https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/api/queries/image_download_api.py
        self.annsetsapi = (
            AnnotatedSectionDataSetsApi()
        )  # https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/api/queries/annotated_section_data_sets_api.py
        self.oapi = (
            OntologiesApi()
        )  # https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/api/queries/ontologies_api.py

        # Get metadata about atlases
        self.atlases = pd.DataFrame(self.oapi.get_atlases_table())
        self.atlases_names = sorted(list(self.atlases["name"].values))

        self.mouse_coronal_atlas_id = int(
            self.atlases.loc[
                self.atlases["name"] == self.mouse_coronal
            ].id.values[0]
        )
        self.mouse_sagittal_atlas_id = int(
            self.atlases.loc[
                self.atlases["name"] == self.mouse_sagittal
            ].id.values[0]
        )
        self.mouse_3D_atlas_id = int(
            self.atlases.loc[self.atlases["name"] == self.mouse3d].id.values[0]
        )

        # Get metadata about products
        self.products = pd.DataFrame(
            send_query(
                "http://api.brain-map.org/api/v2/data/query.json?criteria=model::Product"
            )
        )
        self.mouse_brain_reference_product_id = 12
        self.mouse_brain_ish_data_product_id = 1
        self.products_names = sorted(list(self.products["name"].values))
        self.mouse_products_names = sorted(
            list(
                self.products.loc[self.products.species == "Mouse"][
                    "name"
                ].values
            )
        )

    # UTILS
    def get_atlas_by_name(self, atlas_name):
        """
        Get a brain atlas in the Allen's database given it's name

        :param atlas_name: str with atlas name

        """
        if atlas_name not in self.atlases_names:
            raise ValueError(
                "Available atlases: {}".format(self.atlases_names)
            )
        return int(
            self.atlases.loc[self.atlases["name"] == atlas_name].id.values[0]
        )

    def get_atlas_dataset_id_by_name(self, atlas_name):
        atlas = self.get_atlas_by_name(atlas_name)
        return self.get_atlasimages_by_atlasid(atlas)["data_set_id"].values[0]

    def get_products_by_species(self, species):
        """
        Get all 'products' in the Allen Database for a given species

        :param species: str

        """
        return self.products.loc[self.products.species == species]

    def get_experimentsid_by_productid(self, productid, **kwargs):
        """
        Get the experiment's ID that belong to the same project (product).

        :param productid: int with product ID number
        :param **kwargs: 

        """
        # for more details: https://github.com/AllenInstitute/AllenSDK/blob/master/allensdk/api/queries/image_download_api.py
        return pd.DataFrame(
            self.get_section_data_sets_by_product([productid], **kwargs)
        )

    def get_experimentimages_by_expid(self, expid):
        """
        Get's images that belong to an experiment

        :param expid: int with experiment name. 

        """
        # expid should be a section dataset id
        return pd.DataFrame(self.section_image_query(expid))

    def get_atlasimages_by_atlasid(self, atlasid):
        """
        Get the metadata of images that belong to an atlas. 

        :param atlasid: int with atlas number

        """
        if not isinstance(atlasid, int):
            raise ValueError(
                "Atlas id should be an integer not: {}".format(atlasid)
            )
        return pd.DataFrame(self.atlas_image_query(atlasid))

    def download_images_by_imagesid(
        self,
        savedir,
        imagesids,
        downsample=0,
        annotated=True,
        snames=None,
        atlas_svg=True,
    ):
        """
        Downloads and saves images given a list of images IDs. 


        :param savedir: str, folder in which to save the image
        :param imagesids: list of int with images IDs
        :param downsample: downsample factor, to reduce the image size and resolution (Default value = 0)
        :param annotated: if True the images are overlayed with annotations  (Default value = True)
        :param snames: if True the images are overlayed with the structures names (Default value = None)
        :param atlas_svg: if True fetches the images as SVG, otherwise as PNG (Default value = True)

        """
        savedir = Path(savedir)
        savedir.mkdir(exist_ok=True)

        curdir = Path.cwd()
        chdir(savedir)

        for i, imgid in track(
            enumerate(imagesids),
            total=len(imagesids),
            description="downloading iamges...",
        ):
            if not atlas_svg and not annotated:
                savename = str(imgid) + ".jpg"
            elif not atlas_svg and annotated:
                savename = str(imgid) + "_annotated.jpg"
            else:
                savename = str(imgid) + ".svg"

            if snames is not None:
                sname, ext = savename.split(".")
                savename = (
                    sname + "_sect{}_img{}.".format(snames[i], i + 1) + ext
                )

            if Path(savename).exists():
                continue

            if not atlas_svg and not annotated:
                self.download_section_image(
                    imgid, file_path=savename, downsample=downsample
                )
            elif not atlas_svg and annotated:
                self.download_atlas_image(
                    imgid,
                    file_path=savename,
                    annotation=True,
                    downsample=downsample,
                )
            else:
                self.download_svg(imgid, file_path=savename)

        chdir(curdir)

    def download_images_by_atlasid(
        self, savedir, atlasid, debug=False, **kwargs
    ):
        """
        Downloads all the images that belong to an altlas

        :param savedir: str, folder in which to save the images
        :param atlasid: int, ID of the atlas to use
        :param **kwargs: keyword arguments for self.download_images_by_imagesid

        """
        imgsids = self.get_atlasimages_by_atlasid(atlasid)["id"]

        if debug:  # use fewer images to speed up
            imgsids = imgsids[:2]

        imgs_secs_n = self.get_atlasimages_by_atlasid(atlasid)[
            "section_number"
        ]

        _ = kwargs.pop("snames", None)

        self.download_images_by_imagesid(
            savedir, imgsids, snames=imgs_secs_n, **kwargs
        )



import brainrender
from pathlib import Path
from vedo import (
    Plotter,
    shapes,
    interactive,
    Text2D,
    closePlotter,
    Plane,
    Mesh,
)
from vedo.shapes import Cylinder, Line
import pyinspect as pi
from pyinspect._colors import orange, mocassin, salmon
from brainrender.Utils.scene_utils import (
    get_scene_atlas,
    get_cells_colors_from_metadata,
    make_actor_label,
)
from brainrender.Utils.data_io import (
    load_mesh_from_file,
    load_cells_from_file,
)
from brainrender.ABA.aba_utils import parse_sharptrack
from brainrender.Utils.data_manipulation import (
    return_list_smart,
    make_optic_canula_cylinder,
)
from brainrender.Utils.camera import set_camera
from brainrender.Utils.actors_funcs import get_actor_midpoint, get_actor_bounds
from brainrender.render import Render
from brainrender.Utils.ruler import ruler


class Scene(Render):
    """
        The code below aims to create a scene to which actors can be added or removed, changed etc..
        It also facilitates the interaction with the scene (e.g. moving the camera) and the creation of
        snapshots or animated videos.
        The class Scene is based on the Plotter class of vedo: https://github.com/marcomusy/vedo/blob/master/vedo/plotter.py
        and other classes within the same package.
    """

    verbose = brainrender.VERBOSE

    def __init__(
        self,
        add_root=None,
        verbose=True,
        display_inset=None,
        base_dir=None,
        camera=None,
        screenshot_kwargs={},
        use_default_key_bindings=False,
        title=None,
        atlas=None,
        atlas_kwargs=dict(),
        **kwargs,
    ):
        """

            Creates and manages a Plotter instance

            :param add_root: if False a rendered outline of the whole brain is added to the scene (default value None)
            :param verbose: if False less feedback is printed to screen (default value True)
            :param display_insert: if False the inset displaying the brain's outline is not rendered (but the root is added to the scene) (default value None)
            :param base_dir: path to directory to use for saving data (default value None)
            :param camera: name of the camera parameters setting to use (controls the orientation of the rendered scene)
            :param kwargs: can be used to pass path to individual data folders. See brainrender/Utils/paths_manager.py
            :param screenshot_kwargs: pass a dictionary with keys:
                        - 'folder' -> str, path to folder where to save screenshots
                        - 'name' -> str, filename to prepend to screenshots files
                        - 'format' -> str, 'png', 'svg' or 'jpg'
                        - scale -> float, values > 1 yield higher resultion screenshots
            :param use_default_key_bindings: if True the defualt keybindings from vedo are used, otherwise
                            a custom function that can be used to take screenshots with the parameter above. 
            :param title: str, if a string is passed a text is added to the top of the rendering window as a title
            :param atlas: str, class. Default None. If a string is passed it whould be the name of a valide
                brainglobe_api atlas. Alternative a class object can be passed, this should support the functionality
                expected of an atlas class. 
                if no atlas is passed the allen brain atlas for the adult mouse brain is used. If a string with the atlas
                name is passed it will try to load the corresponding brainglobe atlas.
            :param atlas_kwargs: dictionary used to pass extra arguments to atlas class
        """
        # Get atlas
        self.atlas = get_scene_atlas(atlas, base_dir, atlas_kwargs, **kwargs)

        # Initialise Render
        Render.__init__(
            self,
            verbose,
            display_inset,
            camera,
            screenshot_kwargs,
            use_default_key_bindings,
        )

        # Prepare store for actors added to scene
        self.actors = []
        self.actors_labels = []
        self.store = {}  # in case we need to store some data

        # Add items to scene
        if add_root is None:
            add_root = brainrender.DISPLAY_ROOT
        self.add_root(render=add_root)

        if title is not None:
            self.add_text(title)

        # Placeholder variables
        self.inset = None  # the first time the scene is rendered create and store the inset here
        self.is_rendered = False  # keep track if scene has been rendered

    def __str__(self):
        return f"A `brainrender.scene.Scene` with {len(self)} actors."

    def __repr__(self):
        return f"A `brainrender.scene.Scene` with {len(self)} actors."

    def __add__(self, other):
        if isinstance(other, Mesh):
            self.add_actor(other)
        elif isinstance(other, (Path, str)):
            self.add_from_file(str(other))
        else:
            raise ValueError(f"Can't add object {other} to scene")

    def __iadd__(self, other):
        self.__add__(other)
        return self

    def __len__(self):
        return len(self.actors)

    def __getitem__(self, index):
        return self.actors[index]

    def __del__(self):
        self.close()

    # ---------------------------------------------------------------------------- #
    #                                     Utils                                    #
    # ---------------------------------------------------------------------------- #

    def _get_inset(self, **kwargs):
        """
        Handles the rendering of the inset showing the outline of the whole brain (root) in a corner of the scene.

        :param **kwargs:

        """
        if "plotter" in list(kwargs.keys()):
            self.add_root(render=False, **kwargs)

        if self.display_inset and self.inset is None:
            if self.root is None:
                self.add_root(render=False, **kwargs)
                if self.root is None:
                    print("Could not find root object")
                    return

                self.inset = self.root.clone().scale(0.5)
                self.root = None
                self.actors["root"] = None
            else:
                self.inset = self.root.clone().scale(0.5)

            self.inset.alpha(1)
            self.plotter.showInset(
                self.inset, pos=(0.95, 0.1), draggable=False
            )

    def cut_actors_with_plane(
        self,
        plane,
        actors=None,
        showplane=False,
        returncut=False,
        close_actors=False,
        **kwargs,
    ):
        # Check arguments
        if isinstance(plane, (list, tuple)):
            planes = plane.copy()
        else:
            planes = [plane]

        if actors is None:
            actors = self.actors
        else:
            if not isinstance(actors, (list, tuple)):
                actors = [actors]

        # Loop over each plane
        to_return = []
        for plane in planes:
            # Get the plane actor
            if isinstance(plane, str):
                plane = self.atlas.get_plane_at_point(plane=plane, **kwargs)
            else:
                if not isinstance(plane, Plane):
                    raise ValueError(
                        "The plane arguments should either be a Plane actor or"
                        + "a string with the name of predefined planes."
                        + f" Not: {plane.__type__}"
                    )

            # Show plane
            if showplane:
                self.add_actor(plane)

            # Cut actors
            for actor in actors:
                if actor is None:
                    continue

                try:
                    actor = actor.cutWithPlane(
                        origin=plane.center,
                        normal=plane.normal,
                        returnCut=returncut,
                    )
                except AttributeError:
                    # some rendered objects can't be cut (e.g.text 2d)
                    continue

                if returncut:
                    to_return.append(actor)

                if close_actors:
                    actor.cap()

        if len(to_return) == 1:
            return to_return[0]
        else:
            return to_return

    def list_actors(self):
        actors = pi.Report("Scene actors", accent=salmon)
        for act in self.actors:
            try:
                name = act.name
            except AttributeError:
                pass

            try:
                br_class = act._br_class
            except AttributeError:
                raise ValueError(
                    f'Actor {name} doesnt have a "br_class" attribute!'
                )

            actors.add(f"[{orange}]- {name}[{mocassin}] (type: {br_class})")

        actors.print()

    # ---------------------------------------------------------------------------- #
    #                                POPULATE SCENE                                #
    # ---------------------------------------------------------------------------- #

    def add_root(self, render=True, **kwargs):
        """
        adds the root the scene (i.e. the whole brain outline)

        :param render:  (Default value = True)
        :param **kwargs:

        """
        self.root = self.atlas._get_structure_mesh(
            "root",
            color=brainrender.ROOT_COLOR,
            alpha=brainrender.ROOT_ALPHA,
            **kwargs,
        )

        if self.root is not None:
            self.root.name = "root"
            self.root._br_class = "root"
            self.atlas._root_midpoint = get_actor_midpoint(self.root)
            self.atlas._root_bounds = get_actor_bounds(self.root)

        else:
            print("Could not find a root mesh")
            return None

        if render:
            self.actors.append(self.root)
        elif brainrender.SHOW_AXES:
            # if showing axes, add a transparent root
            # so that scene has right scale
            root = self.root.clone().alpha(0)
            self.actors.append(root)

        return self.root

    def add_brain_regions(self, *args, **kwargs):
        """
            Adds brain regions meshes to scene.
            Check the atlas' method to know how it works
        """
        add_labels = kwargs.pop("add_labels", False)

        allactors = self.atlas.get_brain_regions(
            *args, verbose=self.verbose, **kwargs
        )

        actors = []
        for region, actor in allactors.items():
            if region in [a.name for a in self.actors if isinstance(a, Mesh)]:
                # Avoid inserting again
                continue

            if add_labels:
                self.add_actor_label(actor, region, **kwargs)

            actor.name = region
            actor._br_class = "brain region"
            actors.append(actor)

        self.actors.extend(actors)
        return return_list_smart(actors)

    def add_neurons(self, *args, **kwargs):
        """
        Adds rendered morphological data of neurons reconstructions.
        Check the atlas' method to know how it works
        """
        actors, store = self.atlas.get_neurons(*args, **kwargs)

        if store is not None:
            for n, v in store.items():
                self.store[n] = v

        if isinstance(actors, list):
            for act in actors:
                if isinstance(act, dict):
                    for _act in act.values():
                        if _act is not None:
                            _act.name = "neuron"
                            _act._br_class = "neuron"
                else:
                    act.name = "neuron"
                    act._br_class = "neuron"
                self.actors.extend(list(act.values()))
        else:

            for act in list(actors.values()):
                if act is None:
                    continue
                act.name = "neuron"
                act._br_class = "neuron"
            self.actors.extend(list(actors.values()))
        return actors

    def add_neurons_synapses(self, *args, **kwargs):
        """
        Adds the location of pre or post synapses for a neuron (or list of neurons).
        Check the atlas' method to know how it works. 
        """
        spheres_data, actors = self.atlas.get_neurons_synapses(
            self.store, *args, **kwargs
        )

        for data, kwargs in spheres_data:
            self.add_cells(data, **kwargs)

        for actor in actors:
            actor.name = "synapse"
            actor._br_class = "synapse"
            self.add_actor(actor)

    def add_tractography(self, *args, **kwargs):
        """
        Renders tractography data and adds it to the scene. 
        Check the function definition in ABA for more details
        """

        actors = self.atlas.get_tractography(*args, **kwargs)

        for actor in actors:
            actor.name = "tractography"
            actor._br_class = "tractography"

        self.actors.extend(actors)
        return return_list_smart(actors)

    def add_streamlines(self, *args, **kwargs):
        """
        Render streamline data.
        Check the function definition in ABA for more details
        """
        actors = self.atlas.get_streamlines(*args, **kwargs)
        self.actors.extend(actors)

        for act in actors:
            act.name = "streamlines"
            act._br_class = "streamlines"

        return return_list_smart(actors)

    # -------------------------- General actors/elements ------------------------- #
    def add_actor(self, *actors, store=None):
        """
        Add a vtk actor to the scene

        :param actor:
        :param store: a list to store added actors

        """
        for actor in actors:
            if store is None:
                self.actors.append(actor)
            else:
                store.append(actor)
        return return_list_smart(actors)

    def add_silhouette(self, *actors, lw=1, color="k", **kwargs):
        """
            Given a list of actors it adds a colored silhouette
            to them.
        """
        for actor in actors:
            sil = actor.silhouette(**kwargs).lw(lw).c(color)
            sil.name = "silhouette"
            sil._br_class = "silhouette"
            sil._original_mesh = actor
            self.add_actor(sil)

    def add_from_file(self, *filepaths, **kwargs):
        """
        Add data to the scene by loading them from a file. Should handle .obj, .vtk and .nii files.

        :param filepaths: path to the file. Can pass as many arguments as needed
        :param **kwargs:

        """
        actors = []
        for filepath in filepaths:
            actor = load_mesh_from_file(filepath, **kwargs)
            actor.name = Path(filepath).name
            actor._br_class = Path(filepath).name
            self.actors.append(actor)
            actors.append(actor)
        return return_list_smart(actors)

    def add_sphere_at_point(
        self, pos=[0, 0, 0], radius=100, color="black", alpha=1, **kwargs
    ):
        """
        Adds a shere at a location specified by the user

        :param pos: list of x,y,z coordinates (Default value = [0, 0, 0])
        :param radius: int, radius of the sphere (Default value = 100)
        :param color: color of the sphere (Default value = "black")
        :param alpha: transparency of the sphere (Default value = 1)
        :param **kwargs:
        """
        sphere = shapes.Sphere(
            pos=pos, r=radius, c=color, alpha=alpha, **kwargs
        )
        sphere.name = f"sphere {pos}"
        sphere._br_class = "sphere"
        self.actors.append(sphere)
        return sphere

    def add_cells_from_file(
        self, filepath, hdf_key="hdf", color="red", radius=25, res=7, alpha=1
    ):
        """
        Load location of cells from a file (csv and HDF) and render as spheres aligned to the root mesh.

        :param filepath: str path to file
        :param hdf_key: str (Default value = None)
        :param color: str, color of spheres used to render the cells (Default value = "red")
        :param radius: int, radius of spheres used to render the cells (Default value = 25)
        :param res: int, resolution of spheres used to render the cells (Default value = 3)
        :param alpha: float, transparency of spheres used to render the cells (Default value = 1)

        """
        # Load cells
        cells, name = load_cells_from_file(filepath, hdf_key=hdf_key)

        # Render cells
        cells_actor = self.add_cells(
            cells, color=color, radius=radius, res=res, alpha=alpha
        )
        cells_actor.name = name
        cells_actor._br_class = name
        return cells_actor

    def add_cells(
        self,
        coords,
        color="red",
        color_by_region=False,
        color_by_metadata=None,
        radius=25,
        res=7,
        alpha=1,
        col_names=None,
        regions=None,
        verbose=True,
    ):
        """
        Renders cells given their coordinates as a collection of spheres.

        :param coords: pandas dataframe with x,y,z coordinates
        :param color: str, color of spheres used to render the cells (Default value = "red"). 
                Alternatively a list of colors specifying the color of each cell.
        :param radius: int, radius of spheres used to render the cells (Default value = 25)
        :param res: int, resolution of spheres used to render the cells (Default value = 3)
        :param alpha: float, transparency of spheres used to render the cells (Default value = 1)
        :param color_by_region: bool. If true the cells are colored according to the color of the brain region they are in
        :param color_by_metadata: str, if the name of a column of the coords dataframe is passed, cells are colored according 
                to their value for that column. If color_by_metadata is passed and a dictionary is passed
                to 'color' at the same time, the dictionary will be used to specify the colors used. Therefore
                `color` should map values in the metadata column to colors
        :param regions: if a list of brain regions acronym is passed, only cells in these regions will be added to the scene
        :param col_names: list of strings with names of pandas dataframe columns. If passed it should be a list of 3 columns
                which have the x, y, z coordinates. If not passed, it is assumed that the columns are ['x', 'y', 'z']
        """
        coords_df = coords.copy()  # keep a copy
        if col_names is None:
            col_names = ["x", "y", "z"]

        if regions is not None:
            coords = self.get_cells_in_region(coords, regions)

        coords = [
            [x, y, z]
            for x, y, z in zip(
                coords[col_names[0]].values,
                coords[col_names[1]].values,
                coords[col_names[2]].values,
            )
        ]

        # Update color
        if color_by_region:
            color = self.atlas.get_colors_from_coordinates(coords)
        elif color_by_metadata is not None:
            color = get_cells_colors_from_metadata(
                color_by_metadata, coords_df, color
            )

        # Create actors
        spheres = shapes.Spheres(
            coords, c=color, r=radius, res=res, alpha=alpha
        )
        spheres.name = "cells"
        spheres._br_class = "cells"
        self.actors.append(spheres)

        if verbose:
            print("Added {} cells to the scene".format(len(coords)))

        return spheres

    def add_optic_cannula(
        self, *args, **kwargs,
    ):
        """
            Adds a cylindrical vedo actor to scene to render optic cannulas. By default
            this is a semi-transparent blue cylinder centered on the center of mass of
            a specified target region and oriented vertically. Parameters are specified 
            as keyword arguments.

            :param target_region: str, acronym of target region to extract coordinates
                of implanted fiber. By defualt the fiber will be centered on the center
                of mass of the target region but the offset arguments can be used to
                fine tune the position. Alternative pass a 'pos' argument with AP-DV-ML coords.
            :param pos: list or tuple or np.array with X,Y,Z coordinates. Must have length = 3.
            :param x_offset, y_offset, z_offset: int, used to fine tune the coordinates of 
                the implanted cannula.
            :param **kwargs: used to specify which hemisphere the cannula is and parameters
                of the rendered cylinder: color, alpha, rotation axis...
        """

        # Compute params
        params = make_optic_canula_cylinder(
            self.atlas, self.root, *args, **kwargs
        )

        # Create actor
        cylinder = self.add_actor(Cylinder(**params))
        cylinder.name = "optic cannula"
        cylinder._br_class = "optic cannula"
        return cylinder

    def add_text(
        self, text, pos=8, size=2.5, color="k", alpha=1, font="Montserrat"
    ):
        """
            Adds a 2D text to the scene. Default params are to crate a large black
            text at the top of the rendering window.

            :param text: str with text to write
            :param kwargs: keyword arguments accepted by vedo.shapes.Text2D
        """

        txt = self.add_actor(
            Text2D(text, pos=pos, s=size, c=color, alpha=alpha, font=font)
        )
        return txt

    def add_actor_label(self, actors, labels, **kwargs):
        """
            Adds a 2D text ancored to a point on the actor's mesh
            to label what the actor is

            :param kwargs: key word arguments can be passed to determine 
                    text appearance and location:
                        - size: int, text size. Default 300
                        - color: str, text color. A list of colors can be passed
                                if None the actor's color is used. Default None.
                        - xoffset, yoffset, zoffset: integers that shift the label position
                        - radius: radius of sphere used to denote label anchor. Set to 0 or None to hide. 
        """
        labels = make_actor_label(self.atlas, actors, labels, **kwargs)

        # Add to scene and return
        self.add_actor(*labels, store=self.actors_labels)

        return return_list_smart(labels)

    def add_line_at_point(
        self, point, axis, color="blackboard", lw=3, **kwargs
    ):
        """
            Adds a line oriented on a given axis at a point

            :param point:list or 1d np array with coordinates of point where crosshair is centered
            :param replace_coord: index of the coordinate to replace (i.e. along which axis is the line oriented)
            :param bounds: list of two floats with lower and upper bound for line, determins the extent of the line
            :param kwargs: dictionary with arguments to specify how lines should look like
        """
        # TODO bgspace could be used here
        axis_dict = dict(rostrocaudal=0, dorsoventral=1, mediolateral=2)
        replace_coord = axis_dict[axis]
        bounds = self.atlas._root_bounds[replace_coord]
        # Get line coords
        p0, p1 = point.copy(), point.copy()
        p0[replace_coord] = bounds[0]
        p1[replace_coord] = bounds[1]

        # Create line actor
        line = Line(p0, p1, c=color, lw=lw, **kwargs)
        line.name = f"line through {point}"
        line._br_class = "line"
        return self.add_actor(line)

    def add_crosshair_at_point(
        self, point, show_point=True, line_kwargs={}, point_kwargs={},
    ):
        """
            Add a crosshair (set of orthogonal lines meeting at a point)
            centered on a given point.

            :param point: list or 1d np array with coordinates of point where crosshair is centered
            :param show_point: bool, if True a sphere at the loation of the point is shown
            :param line_kwargs: dictionary with arguments to specify how lines should look like
            :param point_kwargs: dictionary with arguments to specify how the point should look
        """
        return [
            self.add_line_at_point(point, ax, **line_kwargs)
            for ax in ["rostrocaudal", "dorsoventral", "mediolateral"]
        ]

    def add_plane(self, plane, **kwargs):
        """
            Adds one or more planes to the scene.
            For more details on how to build custom planes, check:
            brainrender/atlases/base.py -> Base.get_plane_at_point 
            method.

            :param plane: either a string with the name of one of 
                the predifined planes ['sagittal', 'coronal', 'horizontal'] 
                or an instance of the Plane class from vedo.shapes
        """
        if isinstance(plane, (list, tuple)):
            planes = plane.copy()
        else:
            planes = [plane]

        actors = []
        for plane in planes:
            if isinstance(plane, str):
                plane = self.atlas.get_plane_at_point(plane=plane, **kwargs)
            else:
                if not isinstance(plane, Plane):
                    raise ValueError(
                        "The plane arguments should either be a Plane actor or"
                        + "a string with the name of predefined planes."
                        + f" Not: {plane.__type__}"
                    )

            plane.name = "plane"
            plane._br_class = "plane"
            actors.append(plane)

        self.add_actor(*actors)
        return return_list_smart(actors)

    def add_ruler_from_surface(self, p0, unit_scale=1, axis=1):
        """
            Given a point, this function adds a ruler object showing
            the distance of that point from the brain surface along a
            given axis.

            :param p0: np.array with point coordinates
            :param axis: int, index of the axis along
                    which to measure distance
        """
        # Get point on brain surface
        p1 = p0.copy()
        p1[axis] = 0  # zero the choosen coordinate

        pts = self.root.intersectWithLine(p0, p1)
        surface_point = pts[0]

        # create ruler
        return self.add_actor(
            ruler(surface_point, p0, unit_scale=unit_scale, units="mm",)
        )

    # ----------------------- Application specific methods ----------------------- #

    def add_probe_from_sharptrack(
        self, probe_points_file, points_kwargs={}, name=None, **kwargs
    ):
        """
            Visualises the position of an implanted probe in the brain. 
            Uses the location of points along the probe extracted with SharpTrack
            [https://github.com/cortex-lab/allenCCF].
            It renders the position of points along the probe and a line fit through them.
            Code contributed by @tbslv on github. 

            :param probe_points_file: str, path to a .mat file with probe points coordinates
            :param kwargs: keyword arguments used to specify how the probe and the poitns 
                                should look like (e.g. color, alpha...)
        """
        probe_points_df, points_params, probe, color = parse_sharptrack(
            self.atlas, probe_points_file, name, **kwargs
        )

        spheres = self.add_cells(probe_points_df, **points_params)

        probe.name = "probe"
        probe._br_class = "probe"
        spheres.name = "spheres"
        spheres._br_class = "spheres"

        self.add_actor(spheres, probe)
        return probe, spheres


# ---------------------------------------------------------------------------- #
#                                 OTHER SCENES                                 #
# ---------------------------------------------------------------------------- #


class DualScene:
    """ """

    # A class that manages two scenes to display side by side
    def __init__(self, *args, **kwargs):
        self.scenes = [Scene(*args, **kwargs), Scene(*args, **kwargs)]

    def render(self, _interactive=True):
        """ """
        mv = Plotter(
            N=2,
            axes=4 if brainrender.SHOW_AXES else 0,
            size="full" if brainrender.WHOLE_SCREEN else "auto",
            pos=brainrender.WINDOW_POS,
            bg=brainrender.BACKGROUND_COLOR,
            sharecam=True,
        )

        actors = []
        for scene in self.scenes:
            scene.apply_render_style()
            actors.append(scene.actors)
            mv.add(scene.actors)

        mv.show(
            actors[0],
            at=0,
            zoom=1.15,
            axes=4 if brainrender.SHOW_AXES else 0,
            roll=180,
            interactive=False,
        )
        mv.show(actors[1], at=1, interactive=False)

        if _interactive:
            interactive()

    def close(self):
        closePlotter()


class MultiScene:
    """ """

    def __init__(self, N, scenes=None, *args, **kwargs):
        if scenes is None:
            self.scenes = [Scene(*args, **kwargs) for i in range(N)]
        else:
            if not isinstance(scenes, (list, tuple)):
                raise ValueError("scenes must be a list or a tuple")
            if len(scenes) != N:
                raise ValueError(
                    "Wrong number of scenes passed, it should match N"
                )
            self.scenes = scenes
        self.N = N

    def render(self, _interactive=True, **kwargs):
        """

        :param _interactive:  (Default value = True)
        :param **kwargs:

        """
        camera = kwargs.pop("camera", None)

        for scene in self.scenes:
            scene.apply_render_style()

            if camera is None:
                if scene.atlas.default_camera is None:
                    scene_camera = brainrender.CAMERA
                else:
                    scene_camera = scene.atlas.default_camera
            else:
                if camera:
                    scene_camera = camera
                else:
                    scene_camera = None
            if scene_camera is not None:
                set_camera(scene, scene_camera)

        mv = Plotter(
            N=self.N,
            axes=4 if brainrender.SHOW_AXES else 0,
            size="full" if brainrender.WHOLE_SCREEN else "auto",
            sharecam=True,
            bg=brainrender.BACKGROUND_COLOR,
        )

        actors = []
        for i, scene in enumerate(self.scenes):
            scene.apply_render_style()
            actors.append(scene.actors)
            mv.add(scene.actors)

        for i, scene.actors in enumerate(actors):
            mv.show(scene.actors, at=i, interactive=False)

        print("Rendering complete")
        if _interactive:
            interactive()

    def close(self):
        closePlotter()



""" 
    This tutorial shows how to create and render a brainrender scene with a title on top
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

scene = Scene(title="The thalamus.")

# You can use scene.add_text to add other text elsewhere in the scene.

scene.add_brain_regions(["TH"], alpha=0.4)
scene.render(camera="sagittal")



import numpy as np
from accepts import accepts
from brainrender.colors import (
    getColor,
    colorMap,
    _mapscales_cmaps,
    get_random_colors,
)


@accepts((str, list, dict), (type(None), str, list, np.ndarray, dict, tuple))
def parse_neurons_colors(neurons, color):
    """
        Prepares the color info to render neurons

        :param neurons: str, list, dict. File(s) with neurons data or list of rendered neurons.
        :param color: default None. Can be:
                - None: each neuron is given a random color
                - color: rbg, hex etc. If a single color is passed all neurons will have that color
                - cmap: str with name of a colormap: neurons are colored based on their sequential order and cmap
                - dict: a dictionary specifying a color for soma, dendrites and axon actors, will be the same for all neurons
                - list: a list of length = number of neurons with either a single color for each neuron
                        or a dictionary of colors for each neuron
    """

    N = len(neurons)
    colors = dict(soma=None, axon=None, dendrites=None,)

    # If no color is passed, get random colors
    if color is None:
        cols = get_random_colors(N)
        if not isinstance(cols, list):
            cols = [cols]
        colors = dict(soma=cols, axon=cols, dendrites=cols,)
    else:
        if isinstance(color, str):
            # Deal with a a cmap being passed
            if color in _mapscales_cmaps:
                cols = [
                    colorMap(n, name=color, vmin=-2, vmax=N + 2)
                    for n in np.arange(N)
                ]
                colors = dict(
                    soma=cols.copy(), axon=cols.copy(), dendrites=cols.copy(),
                )

            else:
                # Deal with a single color being passed
                cols = [getColor(color) for n in np.arange(N)]
                colors = dict(
                    soma=cols.copy(), axon=cols.copy(), dendrites=cols.copy(),
                )
        elif isinstance(color, dict):
            # Deal with a dictionary with color for each component
            if "soma" not in color.keys():
                raise ValueError(
                    f"When passing a dictionary as color argument, \
                                            soma should be one fo the keys: {color}"
                )
            dendrites_color = color.pop("dendrites", color["soma"])
            axon_color = color.pop("axon", color["soma"])

            colors = dict(
                soma=[color["soma"] for n in np.arange(N)],
                axon=[axon_color for n in np.arange(N)],
                dendrites=[dendrites_color for n in np.arange(N)],
            )

        elif isinstance(color, (list, tuple)):
            # Check that the list content makes sense
            if len(color) != N:
                raise ValueError(
                    "When passing a list of color arguments, the list length"
                    + f" ({len(color)}) should match the number of neurons ({N})."
                )
            if len(set([type(c) for c in color])) > 1:
                raise ValueError(
                    "When passing a list of color arguments, all list elements"
                    + " should have the same type (e.g. str or dict)"
                )

            if isinstance(color[0], dict):
                # Deal with a list of dictionaries
                soma_colors, dendrites_colors, axon_colors = [], [], []

                for col in color:
                    if "soma" not in col.keys():
                        raise ValueError(
                            f"When passing a dictionary as col argument, \
                                                    soma should be one fo the keys: {col}"
                        )
                    dendrites_colors.append(col.pop("dendrites", col["soma"]))
                    axon_colors.append(col.pop("axon", col["soma"]))
                    soma_colors.append(col["soma"])

                colors = dict(
                    soma=soma_colors,
                    axon=axon_colors,
                    dendrites=dendrites_colors,
                )

            else:
                if isinstance(color, tuple):
                    color = [color]
                # Deal with a list of colors
                colors = dict(
                    soma=color.copy(),
                    axon=color.copy(),
                    dendrites=color.copy(),
                )
        else:
            raise ValueError(
                f"Color argument passed is not valid. Should be a \
                                    str, dict, list or None, not {type(color)}:{color}"
            )

    # Check colors, if everything went well we should have N colors per entry
    for k, v in colors.items():
        if len(v) != N:
            raise ValueError(
                f"Something went wrong while preparing colors. Not all \
                            entries have right length. We got: {colors}"
            )

    return colors



from vedo import settings as vedosettings
from vedo import Plotter, show, closePlotter, buildRulerAxes
import datetime
import warnings
import numpy as np
from pathlib import Path
from pyinspect.classes import Enhanced

import brainrender
from brainrender.Utils.scene_utils import (
    get_scene_camera,
    get_scene_plotter_settings,
)
from brainrender.Utils.camera import (
    check_camera_param,
    set_camera,
    get_camera_params,
)
from rich import print
from pyinspect._colors import mocassin, orange

"""
    The render class expands scene.Scene's functionality
    to take care of the rendering and visualization
    of all actors added a scene. 
"""

mtx = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])


class Render(Enhanced):
    _axes_order_corrected = (
        False  # at first render the axes orders is corrected
    )

    def __init__(
        self,
        verbose,
        display_inset,
        camera,
        screenshot_kwargs,
        use_default_key_bindings,
    ):
        """
            Creates and manages a Plotter instance

            :param add_root: if False a rendered outline of the whole brain is added to the scene (default value None)
            :param verbose: if False less feedback is printed to screen (default value True)
            :param display_inset: if False the inset displaying the brain's outline is not rendered (but the root is added to the scene) (default value None)
            :param camera: name of the camera parameters setting to use (controls the orientation of the rendered scene)
            :param screenshot_kwargs: pass a dictionary with keys:
                        - 'folder' -> str, path to folder where to save screenshots
                        - 'name' -> str, filename to prepend to screenshots files
            :param use_default_key_bindings: if True the defualt keybindings from vedo are used, otherwise
                            a custom function that can be used to take screenshots with the parameter above. 
        """
        Enhanced.__init__(self)

        # Setup a few rendering options
        self.verbose = verbose
        self.display_inset = (
            display_inset
            if display_inset is not None
            else brainrender.DISPLAY_INSET
        )

        if vedosettings.notebookBackend == "k3d":
            self.jupyter = True
        else:
            self.jupyter = False

        if self.display_inset and self.jupyter:
            if self.verbose:
                print(
                    "Setting 'display_inset' to False as this feature is not \
                                available in juputer notebooks"
                )
            self.display_inset = False

        # Camera parameters
        self.camera = get_scene_camera(camera, self.atlas)

        # Create vedo plotter
        self.plotter = Plotter(
            **get_scene_plotter_settings(
                self.jupyter, self.atlas, self.verbose
            )
        )

        if brainrender.AXES_STYLE == 7 and brainrender.SHOW_AXES:
            self.make_custom_axes = True  # to be made at render
        else:
            self.make_custom_axes = False

        # SCreenshots and keypresses variables
        self.screenshots_folder = Path(
            screenshot_kwargs.pop("folder", self.atlas.output_screenshots)
        )
        self.screenshots_name = screenshot_kwargs.pop(
            "name", brainrender.DEFAULT_SCREENSHOT_NAME
        )

        if not use_default_key_bindings:
            self.plotter.keyPressFunction = self.keypress
            self.verbose = False

        if not brainrender.SCREENSHOT_TRANSPARENT_BACKGROUND:
            vedosettings.screenshotTransparentBackground = False
            vedosettings.useFXAA = True

    def _make_custom_axes(self):
        """
            When using `ruler` axes (vedy style 7), we need to 
            customize them a little bit, this function takes care of it. 
        """
        raise NotImplementedError(
            "Currently ony AXES_STYLE=1 is supported, sorry"
        )

        # Get plotter and axes color
        plt = self.plotter
        c = (0.9, 0.9, 0.9)
        if np.sum(plt.renderer.GetBackground()) > 1.5:
            c = (0.1, 0.1, 0.1)

        bounds = [
            item for sublist in self.atlas._root_bounds for item in sublist
        ]
        rulax = buildRulerAxes(
            bounds,
            c=c,
            units="μm",
            xtitle="AP - ",
            ytitle="DV - ",
            ztitle="LR - ",
            precision=1,
            labelRotation=0,
            axisRotation=90,
            xycross=False,
        )
        rulax.UseBoundsOff()
        rulax.PickableOff()
        plt.renderer.AddActor(rulax)
        plt.axes_instances[0] = rulax

        return

    def _correct_axes(self):
        """
            When the scene is first rendered, a transform matrix
            is applied to each actor's points to correct orientation
            mismatches: https://github.com/brainglobe/bg-atlasapi/issues/73
        """
        self._axes_order_corrected = True

        # Flip every actor's orientation
        _silhouettes = []
        for actor in self.actors:
            try:
                _name = actor.name
            except AttributeError:
                """ not all scene objects will have a name """
                continue

            if _name != "silhouette":
                try:
                    actor.applyTransform(mtx).reverse()
                except AttributeError:
                    pass
            else:
                """
                    Silhouettes don't transform properly,
                    we need to re-generate them
                """
                _silhouettes.append(actor)

        for sil in _silhouettes:
            self.actors.pop(self.actors.index(sil))
            self.add_silhouette(sil._original_mesh)
            del sil

    def apply_render_style(self):
        if brainrender.SHADER_STYLE is None:  # No style to apply
            return

        for actor in self.actors:
            if actor is not None:
                try:
                    if brainrender.SHADER_STYLE != "cartoon":
                        actor.lighting(style=brainrender.SHADER_STYLE)
                    else:
                        actor.lighting("off")
                except AttributeError:
                    pass  # Some types of actors such as Text 2D don't have this attribute!

    def render(
        self, interactive=True, video=False, camera=None, zoom=None, **kwargs
    ):
        """
        Takes care of rendering the scene
        """
        self.apply_render_style()

        if not video:
            if (
                not self.jupyter
            ):  # cameras work differently in jupyter notebooks?
                if camera is None:
                    camera = self.camera

                if isinstance(
                    camera, (str, dict)
                ):  # otherwise assume that it's vtk.camera
                    camera = check_camera_param(camera)

                set_camera(self, camera)

            if interactive and self.verbose:
                if not self.jupyter:
                    print(
                        f"\n\n[{mocassin}]Rendering scene.\n   Press [{orange}]'q'[/{orange}] to Quit"
                    )
                elif self.jupyter:
                    print(
                        f"[{mocassin}]The scene is ready to render in your jupyter notebook"
                    )

            self._get_inset()

        if zoom is None and not video:
            zoom = 1.2 if brainrender.WHOLE_SCREEN else 1.5

        # Make mesh labels follow the camera
        if not self.jupyter:
            for txt in self.actors_labels:
                txt.followCamera(self.plotter.camera)

        self.is_rendered = True

        args_dict = dict(
            interactive=interactive,
            zoom=zoom,
            bg=brainrender.BACKGROUND_COLOR,
            axes=self.plotter.axes,
        )

        if video:
            args_dict["offscreen"] = True

        if self.make_custom_axes:
            self._make_custom_axes()
            self.make_custom_axes = False

        # Correct axes orientations
        if not self._axes_order_corrected:
            self._correct_axes()

        show(*self.actors, *self.actors_labels, **args_dict)

    def close(self):
        closePlotter()

    def export_for_web(self, filepath="brexport.html"):
        """
            This function is used to export a brainrender scene
            for hosting it online. It saves an html file that can
            be opened in a web browser to show an interactive brainrender scene
        """
        if not filepath.endswith(".html"):
            raise ValueError("Filepath should point to a .html file")

        # prepare settings
        vedosettings.notebookBackend = "k3d"

        # Create new plotter and save to file
        plt = Plotter()
        plt.add(self.actors)
        plt = plt.show(interactive=False)
        plt.camera[-2] = -1

        if self.verbose:
            print(
                "Ready for exporting. Exporting scenes with many actors might require a few minutes"
            )
        with open(filepath, "w") as fp:
            fp.write(plt.get_snapshot())

        if self.verbose:
            print(
                f"The brainrender scene has been exported for web. The results are saved at {filepath}"
            )

        # Reset settings
        vedosettings.notebookBackend = None
        self.jupyter = False

    # ---------------------------------------------------------------------------- #
    #                               USER INTERACTION                               #
    # ---------------------------------------------------------------------------- #
    def keypress(self, key):
        if key == "s":
            self.take_screenshot()

        elif key == "q":
            self.close()

        elif key == "c":
            print(f"Camera parameters:\n{get_camera_params(scene=self)}")

    def take_screenshot(self):
        if not self.is_rendered:
            print(
                "You need to render the scene before you can take a screenshot"
            )
            return

        if brainrender.SCREENSHOT_TRANSPARENT_BACKGROUND:
            warnings.warn(
                "BRAINRENDER - settings: screenshots are set to have transparent background. Set the parameter 'SCREENSHOT_TRANSPARENT_BACKGROUND' to False if you'd prefer a not transparent background"
            )

        self.screenshots_folder.mkdir(exist_ok=True)

        savename = str(self.screenshots_folder / self.screenshots_name)
        savename += (
            f'_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}' + ".png"
        )

        print(f"\nSaving screenshot at {savename}\n")
        self.plotter.screenshot(filename=savename)
        return savename



from brainrender.scene import Scene
from morphapi.api.mpin_celldb import MpinMorphologyAPI


def test_fish_neurons():
    api = MpinMorphologyAPI()
    # api.download_dataset()
    neurons_ids = api.get_neurons_by_structure(837)[:5]
    neurons = api.load_neurons(neurons_ids)

    neurons = [
        neuron.create_mesh(soma_radius=1, neurite_radius=1)[1]
        for neuron in neurons
    ]

    scene = Scene(atlas="mpin_zfish_1um", add_root=True, camera="sagittal2")
    scene.add_neurons(neurons)
    scene.render(interactive=False)
    scene.close()



"""
    This tutorial shows you how to render efferent mesoscale connectivity
    data from the Allen mouse connectome project as streamlines coloring
    each injection's streamline individually. 

"""

from brainrender.scene import Scene
from brainrender.colors import makePalette


# Start by creating a scene with the allen brain atlas atlas
scene = Scene(title="streamlines")


# Download streamlines data for injections in the CA1 field of the hippocampus
filepaths, data = scene.atlas.download_streamlines_for_region("CA1")


scene.add_brain_regions(["CA1"], use_original_color=True, alpha=0.2)

# you can pass either the filepaths or the data
colors = makePalette(len(data), "salmon", "lightgreen")
scene.add_streamlines(data, color=colors, show_injection_site=False)

scene.render(camera="sagittal", zoom=1)



import brainrender

brainrender.SHADER_STYLE = "cartoon"

from brainrender.scene import Scene

sharptrack_file = "Examples/example_files/sharptrack_probe_points.mat"

scene = Scene(use_default_key_bindings=True)

th = scene.add_brain_regions("TH", alpha=0.2)
th.wireframe()

scene.add_probe_from_sharptrack(sharptrack_file)

scene.render()



"""
    This example shows how to add a 3d object to the scene by loading
    it from file and moving/scaling it to be aligned correctly.
    In this case the object is a .stl file of the mouse skull, 
    but it could be anything, including experimental and recording devices.

    Note that the mouse skull and brain meshes come from different sources
    so that's why they don't match perfectly.
"""


import brainrender

brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 1

from brainrender.scene import Scene

scene = Scene()

# Load skull from file
skull = scene.add_from_file("Examples/example_files/skull.stl")
skull.c("ivory").alpha(1)

# Align skull and brain (scene.root)
skull_com = skull.centerOfMass()
root_com = scene.root.centerOfMass()

skull.origin(skull.centerOfMass())
skull.rotateY(90).rotateX(180)
skull.x(root_com[0] - skull_com[0])
skull.y(root_com[1] - skull_com[1])
skull.z(root_com[2] - skull_com[2])
skull.x(3500)
skull.rotateZ(-25)
skull.y(7800)
skull.scale([1300, 1500, 1200])


# Cut skull actor to show brain inside
scene.cut_actors_with_plane("sagittal", actors=skull)

# Improve looks
scene.add_silhouette(scene.root, lw=3)
scene.add_silhouette(skull, lw=3)

scene.render()



"""
    This example shows you how to crate a multi scene
    visualisation in brainrender
"""

from brainrender.scene import MultiScene, Scene

# Create 3 scenes
scene1 = Scene()
scene1.add_brain_regions("MOs")

scene2 = Scene(add_root=False)
scene2.add_brain_regions(["TH", "VAL"])

scene3 = Scene()
scene3.add_brain_regions("CA1")


# Create and render a multiscene
multi = MultiScene(3, scenes=[scene1, scene2, scene3])
multi.render()



from vedo import Mesh, Plane, merge
from morphapi.morphology.morphology import Neuron
import numpy as np
import os
from bg_atlasapi.bg_atlas import BrainGlobeAtlas
from pyinspect.classes import Enhanced

import brainrender
from brainrender.Utils.paths_manager import Paths
from brainrender.atlases.aba import ABA
from brainrender.Utils.data_io import load_mesh_from_file
from brainrender.Utils import actors_funcs
from brainrender.Utils.data_manipulation import (
    return_list_smart,
    return_dict_smart,
    flatten,
)
from brainrender.colors import check_colors
from brainrender.Utils.atlases_utils import parse_neurons_colors
from brainrender.morphology.utils import get_neuron_actors_with_morphapi


class Atlas(BrainGlobeAtlas, Paths, ABA, Enhanced):
    default_camera = None

    def __init__(self, atlas_name, *args, base_dir=None, **kwargs):
        # Create brainglobe atlas
        BrainGlobeAtlas.__init__(
            self, *args, atlas_name=atlas_name, print_authors=False, **kwargs
        )

        # Add brainrender paths
        Paths.__init__(self, base_dir=base_dir, **kwargs)

        Enhanced.__init__(self)

        # If it's a mouse atlas, add extra functionality
        if "Mus musculus" == self.metadata["species"]:
            ABA.__init__(self)

    # ---------------------------------------------------------------------------- #
    #                                    METHODS                                   #
    # ---------------------------------------------------------------------------- #
    # ---------------------------------- Planes ---------------------------------- #
    # functions to create oriented planes that can be used to slice actors etc
    def get_plane_at_point(
        self,
        pos=None,
        norm=None,
        plane=None,
        sx=None,
        sy=None,
        color="lightgray",
        alpha=0.25,
        **kwargs,
    ):
        """ 
            Returns a plane going through a point at pos, oriented 
            orthogonally to the vector norm and of width and height
            sx, sy. 

            :param pos: 3-tuple or list with x,y,z, coords of point the plane goes through
            :param norm: 3-tuple with plane's normal vector (optional)
            :param sx, sy: int, width and height of the plane
            :param plane: "sagittal", "horizontal", or "frontal"
            :param color, alpha: plane color and transparency
        """
        axes_pairs = dict(sagittal=(0, 1), horizontal=(2, 0), frontal=(2, 1))

        # Get position
        if pos is None:
            pos = self._root_midpoint
        elif not isinstance(pos, (list, tuple)) or not len(pos) == 3:
            raise ValueError(f"Invalid pos argument: {pos}")

        # Get normal if one is not given
        if norm is None:
            try:
                norm = self.space.plane_normals[plane]
            except KeyError:
                raise ValueError(
                    f"Could not find normals for plane {plane}. Atlas space {self.space} provides these normals: {self.spacel.plane_normals}"
                )

        # Get plane width and height
        idx_pair = (
            axes_pairs[plane]
            if plane is not None
            else axes_pairs["horizontal"]
        )
        wh = [float(np.diff(self._root_bounds[i])) * 1.2 for i in idx_pair]
        if sx is None:
            sx = wh[0]
        if sy is None:
            sy = wh[1]

        # return plane
        return Plane(pos=pos, normal=norm, sx=sx, sy=sy, c=color, alpha=alpha)

    # ----------------------------------- Misc ----------------------------------- #
    def get_region_CenterOfMass(
        self, regions, unilateral=True, hemisphere="right"
    ):
        """
        Get the center of mass of the 3d mesh of one or multiple brain regions.

        :param regions: str, list of brain regions acronyms
        :param unilateral: bool, if True, the CoM is relative to one hemisphere (Default value = True)
        :param hemisphere: str, if unilteral=True, specifies which hemisphere to use ['left' or 'right'] (Default value = "right")
        :returns: coms = {list, dict} -- [if only one regions is passed, then just returns the CoM coordinates for that region.
                                If a list is passed then a dictionary is returned. ]
        """

        if not isinstance(regions, list):
            regions = [regions]

        coms = {}
        for region in regions:
            # Check if input is an actor or if we need to load one
            if isinstance(region, Mesh):
                mesh = region
            else:
                # load mesh corresponding to brain region
                if unilateral:
                    mesh = self.get_region_unilateral(
                        region, hemisphere="left"
                    )
                else:
                    mesh = self._get_structure_mesh(region)
            com = mesh.centerOfMass()

            #  if using right hemisphere, mirror COM
            if unilateral and hemisphere.lower() == "right":
                com = self.mirror_point_across_hemispheres(com)

            coms[region] = com

        # return data
        return return_dict_smart(coms)

    # ---------------------------------------------------------------------------- #
    #                      METHODS SUPPORTING SCENE POPULATION                     #
    # ---------------------------------------------------------------------------- #
    def _get_structure_mesh(self, acronym, **kwargs):
        obj_path = self._get_from_structure(acronym, "mesh_filename")
        return load_mesh_from_file(obj_path, **kwargs)

    def get_brain_regions(
        self,
        *brain_regions,
        add_labels=False,
        colors=None,
        use_original_color=True,
        alpha=None,
        hemisphere=None,
        verbose=False,
        **kwargs,
    ):

        """
                Gets brain regions meshes for rendering
                Many parameters can be passed to specify how the regions should be rendered.
                To treat a subset of the rendered regions, specify which regions are VIP. 
                Use the kwargs to specify more detailes on how the regins should be rendered (e.g. wireframe look)

                :param brain_regions: str list of acronyms of brain regions
                :param colors: str, color of rendered brian regions (Default value = None)
                :param use_original_color: bool, if True, the allen's default color for the region is used.  (Default value = False)
                :param alpha: float, transparency of the rendered brain regions (Default value = None)
                :param hemisphere: str (Default value = None)
                :param add_labels: bool (default False). If true a label is added to each regions' actor. The label is visible when hovering the mouse over the actor
                :param **kwargs: used to determine a bunch of thigs, including the look and location of lables from scene.add_labels
            """
        # Parse arguments
        if alpha is None:
            alpha = brainrender.DEFAULT_STRUCTURE_ALPHA

        # Flatten brain regions argument
        if isinstance(brain_regions[0], list):
            brain_regions = flatten(brain_regions)

        # check the colors input is correct
        if colors is not None:
            if isinstance(colors, (list, tuple)):
                if not len(colors) == len(brain_regions):
                    raise ValueError(
                        "when passing colors as a list, the number of colors must match the number of brain regions"
                    )
                for col in colors:
                    if not check_colors(col):
                        raise ValueError(
                            "Invalide colors in input: {}".format(col)
                        )
            else:
                if not check_colors(colors):
                    raise ValueError(
                        "Invalide colors in input: {}".format(colors)
                    )
                colors = [colors for i in range(len(brain_regions))]

        # loop over all brain regions
        actors = {}
        for i, region in enumerate(brain_regions):
            if verbose:
                print("Rendering: ({})".format(region))

            # get the structure and check if we need to download the object file
            if region not in self.lookup_df.acronym.values:
                print(
                    f"The region {region} doesn't seem to belong to the atlas being used: {self.atlas_name}. Skipping"
                )
                continue

            # Get path to .obj file
            obj_file = str(self.meshfile_from_structure(region))

            # check which color to assign to the brain region
            if use_original_color:
                color = [
                    x / 255
                    for x in self._get_from_structure(region, "rgb_triplet")
                ]
            else:
                if colors is None:
                    color = brainrender.DEFAULT_STRUCTURE_COLOR
                elif isinstance(colors, list):
                    color = colors[i]
                else:
                    color = colors

            # Load the object file as a mesh and store the actor
            if hemisphere is not None:
                if (
                    hemisphere.lower() == "left"
                    or hemisphere.lower() == "right"
                ):
                    obj = self.get_region_unilateral(
                        region, hemisphere=hemisphere, color=color, alpha=alpha
                    )
                else:
                    raise ValueError(
                        f"Invalid hemisphere argument: {hemisphere}"
                    )
            else:
                obj = load_mesh_from_file(obj_file, color=color, alpha=alpha)

            if obj is not None:
                actors_funcs.edit_actor(obj, **kwargs)
                actors[region] = obj
            else:
                print(
                    f"Something went wrong while loading mesh data for {region}, skipping."
                )

        return actors

    def get_neurons(
        self,
        neurons,
        color=None,
        display_axon=True,
        display_dendrites=True,
        alpha=1,
        neurite_radius=None,
        soma_radius=None,
        use_cache=True,
    ):
        """
        Gets rendered morphological data of neurons reconstructions
        Accepts neurons argument as:
            - file(s) with morphological data
            - vedo mesh actor(s) of entire neurons reconstructions
            - dictionary or list of dictionary with actors for different neuron parts

        :param neurons: str, list, dict. File(s) with neurons data or list of rendered neurons.
        :param display_axon, display_dendrites: if set to False the corresponding neurite is not rendered
        :param color: default None. Can be:
                - None: each neuron is given a random color
                - color: rbg, hex etc. If a single color is passed all neurons will have that color
                - cmap: str with name of a colormap: neurons are colored based on their sequential order and cmap
                - dict: a dictionary specifying a color for soma, dendrites and axon actors, will be the same for all neurons
                - list: a list of length = number of neurons with either a single color for each neuron
                        or a dictionary of colors for each neuron
        :param alpha: float in range 0,1. Neurons transparency
        :param neurite_radius: float > 0 , radius of tube actor representing neurites
        :param use_cache: bool, if True a cache is used to avoid having to crate a neuron's mesh anew, otherwise a new mesh is created
        """

        if not isinstance(neurons, (list, tuple)):
            neurons = [neurons]

        # ---------------------------------- Render ---------------------------------- #
        _neurons_actors = []
        for neuron in neurons:
            neuron_actors = {"soma": None, "dendrites": None, "axon": None}

            # Deal with neuron as filepath
            if isinstance(neuron, str):
                if os.path.isfile(neuron):
                    if neuron.endswith(".swc"):
                        neuron_actors, _ = get_neuron_actors_with_morphapi(
                            swcfile=neuron,
                            neurite_radius=neurite_radius,
                            soma_radius=soma_radius,
                            use_cache=use_cache,
                        )
                    else:
                        raise NotImplementedError(
                            "Currently we can only parse morphological reconstructions from swc files"
                        )
                else:
                    raise ValueError(
                        f"Passed neruon {neuron} is not a valid input. Maybe the file doesn't exist?"
                    )

            # Deal with neuron as single actor
            elif isinstance(neuron, Mesh):
                # A single actor was passed, maybe it's the entire neuron
                neuron_actors["soma"] = neuron  # store it as soma
                pass

            # Deal with neuron as dictionary of actor
            elif isinstance(neuron, dict):
                neuron_actors["soma"] = neuron.pop("soma", None)
                neuron_actors["axon"] = neuron.pop("axon", None)

                # Get dendrites actors
                if (
                    "apical_dendrites" in neuron.keys()
                    or "basal_dendrites" in neuron.keys()
                ):
                    if "apical_dendrites" not in neuron.keys():
                        neuron_actors["dendrites"] = neuron["basal_dendrites"]
                    elif "basal_dendrites" not in neuron.keys():
                        neuron_actors["dendrites"] = neuron["apical_dendrites"]
                    else:
                        neuron_actors["dendrites"] = merge(
                            neuron["apical_dendrites"],
                            neuron["basal_dendrites"],
                        )
                else:
                    neuron_actors["dendrites"] = neuron.pop("dendrites", None)

            # Deal with neuron as instance of Neuron from morphapi
            elif isinstance(neuron, Neuron):
                neuron_actors, _ = get_neuron_actors_with_morphapi(
                    neuron=neuron,
                    neurite_radius=neurite_radius,
                    use_cache=use_cache,
                )
            # Deal with other inputs
            else:
                raise ValueError(
                    f"Passed neuron {neuron} is not a valid input"
                )

            # Check that we don't have anything weird in neuron_actors
            for key, act in neuron_actors.items():
                if act is not None:
                    if not isinstance(act, Mesh):
                        raise ValueError(
                            f"Neuron actor {key} is {type(act)} but should be a vedo Mesh. Not: {act}"
                        )

            if not display_axon:
                neuron_actors["axon"] = None
            if not display_dendrites:
                neuron_actors["dendrites"] = None
            _neurons_actors.append(neuron_actors)

        # Color actors
        colors = parse_neurons_colors(neurons, color)
        for n, neuron in enumerate(_neurons_actors):
            if neuron["axon"] is not None:
                neuron["axon"].c(colors["axon"][n])
                neuron["axon"].name = "neuron-axon"
            if neuron["soma"] is not None:
                neuron["soma"].c(colors["soma"][n])
                neuron["soma"].name = "neuron-soma"
            if neuron["dendrites"] is not None:
                neuron["dendrites"].c(colors["dendrites"][n])
                neuron["dendrites"].name = "neuron-dendrites"

        # Return
        return return_list_smart(_neurons_actors), None

    # ---------------------------------------------------------------------------- #
    #                                     UTILS                                    #
    # ---------------------------------------------------------------------------- #

    def get_region_unilateral(
        self, region, hemisphere="both", color=None, alpha=None
    ):
        """
        Regions meshes are loaded with both hemispheres' meshes by default.
        This function splits them in two.

        :param region: str, actors of brain region
        :param hemisphere: str, which hemisphere to return ['left', 'right' or 'both'] (Default value = "both")
        :param color: color of each side's mesh. (Default value = None)
        :param alpha: transparency of each side's mesh.  (Default value = None)

        """
        if color is None:
            color = brainrender.ROOT_COLOR
        if alpha is None:
            alpha = brainrender.ROOT_ALPHA
        bilateralmesh = self._get_structure_mesh(region, c=color, alpha=alpha)

        if bilateralmesh is None:
            print(f"Failed to get mesh for {region}, returning None")
            return None

        com = (
            bilateralmesh.centerOfMass()
        )  # this will always give a point that is on the midline
        right = bilateralmesh.cutWithPlane(origin=com, normal=(0, 0, 1))

        # left is the mirror right # WIP
        com = self.get_region_CenterOfMass("root", unilateral=False)[2]
        left = actors_funcs.mirror_actor_at_point(right.clone(), com, axis="x")

        if hemisphere == "both":
            return left, right
        elif hemisphere == "left":
            return left
        else:
            return right

    def mirror_point_across_hemispheres(self, point):
        delta = point[2] - self._root_midpoint[2]
        point[2] = self._root_midpoint[2] - delta
        return point

    def get_colors_from_coordinates(self, p0):
        """
            Given a point or a list of points returns a list of colors where
            each item is the color of the brain region each point is in
        """
        if isinstance(p0[0], (float, int)):
            p0 = [p0]

        resolution = int(self.metadata["resolution"][0])

        colors = []
        for point in p0:
            point = np.round(np.array(point) / resolution).astype(int)
            try:
                struct = self.structure_from_coords(point, as_acronym=True)
            except KeyError:  # couldn't find any region
                struct = "root"
            colors.append(self._get_from_structure(struct, "rgb_triplet"))
        return return_list_smart(colors)



""" 
    This tutorial shows how download and rendered afferent mesoscale projection data
    using the AllenBrainAtlas (ABA) and Scene classes
"""
import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene(title="tractography")

# Get the center of mass of the region of interest
p0 = scene.atlas.get_region_CenterOfMass("ZI")

# Get projections to that point
tract = scene.atlas.get_projection_tracts_to_target(p0=p0)

# Add the brain regions and the projections to it
scene.add_brain_regions(["ZI"], alpha=0.4, use_original_color=True)
scene.add_tractography(tract, color_by="target_region")

scene.render()



"""
This example shows how to crate an animate scene where
over a number of frames:
    - Different sets of streamlines data from the mouse connectome project
    - The camera moves around 
"""


import brainrender

brainrender.WHOLE_SCREEN = False
brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 0.1


from brainrender.scene import Scene
import time
import numpy as np
from random import choices
from rich.progress import track


from brainrender.Utils.camera import (
    top_camera,
    buildcam,
    sagittal_camera,
)

# --------------------------------- Variables -------------------------------- #
minalpha = 0  # transparency of background neurons
darkcolor = "lightgray"  # background neurons color
lightcolor = "lawngreen"  # highlighted neurons color

N_FRAMES = 250
N_streamlines = 5  # number of streamlines to show in total, if -1 all streamlines are shown but it might take a while to render them at first
N_streamlines_in_frame = (
    2  # number of streamlines to be highlighted in a given frame
)
N_frames_for_change = 15  # every N frames which streamlines are shown changes

# Variables to specify camera position at each frame
zoom = np.linspace(1, 1.5, N_FRAMES)
frac = np.zeros_like(
    zoom
)  # for camera transition, interpolation value between cameras
frac[:150] = np.linspace(0, 1, 150)
frac[150:] = np.linspace(1, 0, len(frac[150:]))

# ------------------------------- Create scene ------------------------------- #
scene = Scene(display_inset=True, use_default_key_bindings=True)

filepaths, data = scene.atlas.download_streamlines_for_region("TH")
tracts = scene.add_streamlines(
    data, color="darkseagreen", show_injection_site=False
)

scene.add_brain_regions(["TH"], alpha=0.2)

# Make all streamlines background
for mesh in tracts:
    mesh.alpha(minalpha)
    mesh.color(darkcolor)


# Create new cameras
cam1 = buildcam(sagittal_camera)


cam2 = buildcam(top_camera)

cam3 = buildcam(
    dict(
        position=[1862.135, -4020.792, -36292.348],
        focal=[6587.835, 3849.085, 5688.164],
        viewup=[0.185, -0.97, 0.161],
        distance=42972.44,
        clipping=[29629.503, 59872.10],
    )
)

# Iniziale camera position
startcam = scene.plotter.moveCamera(cam1, cam2, frac[0])


# ------------------------------- Create frames ------------------------------ #
# Create frames
prev_streamlines = []
for step in track(
    np.arange(N_FRAMES), total=N_FRAMES, description="Generating frames..."
):
    if step % N_frames_for_change == 0:  # change neurons every N framse

        # reset neurons from previous set of neurons
        for mesh in prev_streamlines:
            mesh.alpha(minalpha)
            mesh.color(darkcolor)
        prev_streamlines = []

        # highlight new neurons
        streamlines = choices(tracts, k=N_streamlines_in_frame)
        for n, mesh in enumerate(streamlines):
            # color = colorMap(n, 'Reds', vmin=-2, vmax=N_streamlines_in_frame+3)
            mesh.alpha(0.7)
            mesh.color("orangered")
            prev_streamlines.append(mesh)

    # Move scene camera between 3 cameras
    if step < 150:
        scene.plotter.moveCamera(cam1, cam2, frac[step])
    else:
        scene.plotter.moveCamera(cam3, cam2, frac[step])

    # Update rendered window
    time.sleep(0.1)
    scene.render(zoom=zoom[step], interactive=False, video=True)



import sys
import os

"""
    Class to create and store paths to a number of folders uesed to save/load data
"""


# Default paths for Data Folders (store stuff like object meshes, neurons morphology data etc)
default_paths = dict(
    # OUTPUT Folders
    output_screenshots="Output/Screenshots",
    output_videos="Output/Videos",
    output_data="Output/Data",
    # Allen caches
    mouse_connectivity_cache="Data/ABA/MCC",
    annotated_volume_fld="Data/ABA",
    mouse_connectivity_volumetric="Data/ABA/Volumetric",
    mouse_connectivity_volumetric_cache="Data/ABA/Volumetric/cache",
    gene_expression_cache="Data/ABA/Volumetric/GeneExpressionCache",
    # Streamlines cache
    streamlines_cache="Data/Streamlines",
    ibdb_meshes_folder="Data/InsectsDBs",
)


class Paths:
    _folders = list(default_paths.keys())

    def __init__(self, base_dir=None, **kwargs):
        """
        Parses a YAML file to get data folders paths. Stores paths to a number of folders used throughtout brainrender. 
        Other classes (e.g. brainrender.Scene) subclass Paths.
        
        :param base_dir: str with path to directory to use to save data. If none the user's base directiry is used. 
        :param kwargs: use the name of a folder as key and a path as argument to specify the path of individual subfolders
        """
        # Get and make base directory
        if base_dir is None:
            user_dir = os.path.expanduser("~")
            if not os.path.isdir(user_dir):
                raise FileExistsError(
                    "Could not find user base folder (to save brainrender data). Platform: {}".format(
                        sys.platform
                    )
                )
            self.base_dir = os.path.join(user_dir, ".brainrender")
        else:
            self.base_dir = base_dir

        if not os.path.isdir(self.base_dir):
            os.mkdir(self.base_dir)

        for fld_name in self._folders:
            # Check if user provided a path for this folder, otherwise use default
            fld_path = kwargs.pop(fld_name, default_paths[fld_name])

            # Make complete path and save it as an attribute of this class
            path = os.path.join(self.base_dir, fld_path)

            # Create folder if it doesn't exist
            if not os.path.isdir(path):
                print("Creating folder at: {}".format(path))
                os.makedirs(path)
            self.__setattr__(fld_name, path)



"""
    This example shows how a brainrender scene can be exorted to a .html file.
    You can then open the html file in a browser to render an interactive brainrender scene
"""

import brainrender

brainrender.SHADER_STYLE = "cartoon"
from brainrender.scene import Scene

# Create a scene
scene = Scene()

# Add the whole thalamus in gray
scene.add_brain_regions(["TH"], alpha=0.15)


scene.export_for_web()  # <- you can pass a  filepath to specify where to save the scene



from pathlib import Path
from brainrender.Utils.webqueries import connected_to_internet


def fail_on_no_connection(func):
    if not connected_to_internet():
        raise ConnectionError("No internet connection found.")

    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    return inner


def check_file_exists(func):
    def inner(*args, **kwargs):
        if not Path(args[0]).exists():
            raise FileNotFoundError(f"File {args[0]} not found")
        return func(*args, **kwargs)

    return inner



import requests

# ----------------------------- Internet queries ----------------------------- #


def connected_to_internet(url="http://www.google.com/", timeout=5):
    """
        Check that there is an internet connection

        :param url: url to use for testing (Default value = 'http://www.google.com/')
        :param timeout:  timeout to wait for [in seconds] (Default value = 5)
    """

    try:
        _ = requests.get(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        print("No internet connection available.")
    return False


def send_query(query_string, clean=False):
    """
    Send a query/request to a website

    :param query_string: string with query content
    :param clean:  (Default value = False)

    """
    response = requests.get(query_string)
    if response.ok:
        if not clean:
            return response.json()["msg"]
        else:
            return response.json()
    else:
        raise ValueError("Invalide query string: {}".format(query_string))


def request(url):
    """
    Sends a request to a url

    :param url: 

    """
    if not connected_to_internet():
        raise ConnectionError(
            "You need to have an internet connection to send requests."
        )
    response = requests.get(url)
    if response.ok:
        return response
    else:
        exception_string = "URL request failed: {}".format(response.reason)
    raise ValueError(exception_string)



from brainrender import cli
from click.testing import CliRunner
import pytest


def test_cli():
    runner = CliRunner()
    runner.invoke(cli.main, ["TH", "-d"])


def test_cli_args():
    runner = CliRunner()
    runner.invoke(cli.main, ["TH", "STR", "c", "-d"])


@pytest.mark.slow
def test_cli_atlas():
    runner = CliRunner()
    runner.invoke(cli.main, ["TH", "-a", "allen_human_500um", "-d"])



import click
from brainrender.scene import Scene
import brainrender


@click.command()
@click.option("-a", "--atlas", default=None, help="atlas name")
@click.option("-c", "--cartoon", is_flag=True, help="cartoon look")
@click.option("-d", "--debug", is_flag=True)
@click.option("-f", "--file", help="path to .obj or .h5 file")
@click.argument("regions", nargs=-1)
def main(regions, atlas=None, cartoon=False, debug=False, file=None):
    # Set look
    if cartoon:
        brainrender.SHADER_STYLE = "cartoon"

    # Create scene
    scene = Scene(atlas=atlas)

    # Add brain regions
    if regions is not None and len(regions) > 0:
        acts = scene.add_brain_regions(list(regions))

        # Add silhouettes
        if cartoon:
            if isinstance(acts, list):
                scene.add_silhouette(*acts)
            else:
                scene.add_silhouette(acts)

    # Add data from file
    if file is not None:
        if file.endswith(".h5"):
            scene.add_cells_from_file(file)
        else:
            try:
                scene.add_from_file(file)
            except Exception as e:
                raise ValueError(
                    f"Failed to load data from file onto scene: {file}\n{e}"
                )

    # If debug set interactive = Off and close scene
    if not debug:
        interactive = True
    else:
        interactive = False

    # Render and close
    scene.render(interactive=interactive)

    if debug:
        scene.close()



import numpy as np
from vedo import Volume
import pandas as pd
import os
import sys

from brainrender.ABA.gene_expression.ge_utils import (
    check_gene_cached,
    load_cached_gene,
    download_and_cache,
)
from brainrender.Utils.paths_manager import Paths
from brainrender.Utils.webqueries import request
from brainrender.Utils.decorators import fail_on_no_connection


class GeneExpressionAPI(Paths):
    voxel_size = 200  # um
    grid_size = [58, 41, 67]  # number of voxels along each direction

    all_genes_url = (
        "http://api.brain-map.org/api/v2/data/query.json?criteria="
        + "model::Gene,"
        + "rma::criteria,products[abbreviation$eq'DevMouse'],"
        + "rma::options,[tabular$eq'genes.id','genes.acronym+as+gene_symbol','genes.name+as+gene_name',"
        + "'genes.entrez_id+as+entrez_gene_id','genes.homologene_id+as+homologene_group_id'],"
        + "[order$eq'genes.acronym']"
        + "&num_rows=all&start_row=0"
    )

    gene_experiments_url = (
        "http://api.brain-map.org/api/v2/data/query.json?criteria=model::SectionDataSet,"
        + "rma::criteria,[failed$eq'false'],products[abbreviation$eq'Mouse'],genes[acronym$eq'-GENE_SYMBOL-']"
    )

    download_url = "http://api.brain-map.org/grid_data/download/EXP_ID?include=energy,intensity,density"

    def __init__(self, base_dir=None, **kwargs):
        super().__init__(base_dir, **kwargs)

        # Get metadata about all available genes
        self.genes = None  # when necessary gene data can be downloaded with self.get_all_genes

    @fail_on_no_connection
    def get_all_genes(self):
        """
            Download metadata about all the genes available in the Allen gene expression dataset
        """
        res = request(self.all_genes_url)
        return pd.DataFrame(res.json()["msg"])

    def get_gene_id_by_name(self, gene_name):
        if self.genes is None:
            self.genes = self.get_all_genes()

        if gene_name not in self.genes.gene_symbol.values:
            print(
                f"Gene name {gene_name} doesnt appear in the genes dataset, nothing to return\n"
                + "You can search for you gene here: https://mouse.brain-map.org/"
            )
            return None
        else:
            return int(
                self.genes.loc[self.genes.gene_symbol == gene_name].id.values[
                    0
                ]
            )

    def get_gene_symbol_by_id(self, gene_id):
        if self.genes is None:
            self.genes = self.get_all_genes()

        if str(gene_id) not in self.genes.id.values:
            print(
                f"Gene id {gene_id} doesnt appear in the genes dataset, nothing to return\n"
                + "You can search for you gene here: https://mouse.brain-map.org/"
            )
            return None
        else:
            return self.genes.loc[
                self.genes.id == str(gene_id)
            ].gene_symbol.values[0]

    @fail_on_no_connection
    def get_gene_experiments(self, gene_symbol):
        """
            Given a gene_symbol it returns the list of ISH
            experiments for this gene

            :param gene_symbol: str, self.genes.gene_symbol
        """
        if not isinstance(gene_symbol, str):
            if isinstance(gene_symbol, int):  # it's an ID, get symbol
                gene_symbol = self.get_gene_symbol_by_id(gene_symbol)
                if gene_symbol is None:
                    raise ValueError("Invalid gene_symbol argument")
            else:
                raise ValueError("Invalid gene_symbol argument")

        url = self.gene_experiments_url.replace("-GENE_SYMBOL-", gene_symbol)
        data = request(url).json()["msg"]

        if not len(data):
            print(f"No experiment found for gene {gene_symbol}")
            return None
        else:
            return [d["id"] for d in data]

    @fail_on_no_connection
    def download_gene_data(self, gene):
        """
            Downloads a gene's data from the Allen Institute 
            Gene Expression dataset and saves to cache. 
            See: http://help.brain-map.org/display/api/Downloading+3-D+Expression+Grid+Data

            :param gene: int, the gene_id for the gene being downloaded.
        """
        if self.genes is None:
            self.genes = self.get_all_genes()
        if str(gene) not in self.genes.id.values:
            raise ValueError(
                f"The gened {gene} is not in the list of available genes"
            )

        # Get the gene's experiment id
        gene_symbol = self.genes.loc[
            self.genes.id == str(gene)
        ].gene_symbol.values[0]
        exp_ids = self.get_gene_experiments(gene_symbol)

        if exp_ids is None:
            return

        # download experiment data
        for eid in exp_ids:
            print(f"Downloading data for {gene_symbol} - experiment: {eid}")
            url = self.download_url.replace("EXP_ID", str(eid))
            download_and_cache(
                url, os.path.join(self.gene_expression_cache, f"{gene}-{eid}")
            )

    def get_gene_data(self, gene, exp_id, metric="energy"):
        """
            Given a list of gene ids
        """
        if not isinstance(gene, int):
            raise ValueError("Gene id should be an integer")
        if not isinstance(exp_id, int):
            raise ValueError("Expression id should be an integer")

        # Check if gene-experiment cached
        cache = check_gene_cached(self.gene_expression_cache, gene, exp_id)

        if not cache:  # then download it
            self.download_gene_data(gene)
            cache = check_gene_cached(self.gene_expression_cache, gene, exp_id)
            if not cache:
                raise ValueError(
                    "Something went wrong and data were not cached"
                )

        # Load from cache
        data = load_cached_gene(cache, metric, self.grid_size)

        if sys.platform == "darwin":
            data = data.T

        return data

    def griddata_to_volume(
        self, griddata, min_quantile=None, min_value=None, **kwargs
    ):
        """
            Takes a 3d numpy array with volumetric gene expression
            and returns a vedo.Volume.isosurface actor.
            The isosurface needs a lower bound threshold, this can be
            either a user defined hard value (min_value) or the value
            corresponding to some percentile of the gene expression data.

            :param griddata: np.ndarray, 3d array with gene expression data
            :param min_quantile: float, percentile for threshold
            :param min_value: float, value for threshold
        """
        # Check inputs
        if not isinstance(griddata, np.ndarray):
            raise ValueError("Griddata should be a numpy array")
        if not len(griddata.shape) == 3:
            raise ValueError("Griddata should be a 3d array")

        # Get threshold
        if min_quantile is None and min_value is None:
            th = 0
        elif min_value is not None:
            if not isinstance(min_value, (int, float)):
                raise ValueError("min_values should be a float")
            th = min_value
        else:
            if not isinstance(min_quantile, (float, int)):
                raise ValueError(
                    "min_values should be a float in range [0, 1]"
                )
            if 0 > min_quantile or 100 < min_quantile:
                raise ValueError(
                    "min_values should be a float in range [0, 100]"
                )
            th = np.percentile(griddata.ravel(), min_quantile)

        # Create mesh
        cmap = kwargs.pop("cmap", None)
        actor = Volume(
            griddata,
            spacing=[self.voxel_size, self.voxel_size, self.voxel_size],
        )
        actor = actor.legosurface(vmin=th, cmap=cmap)
        return actor



"""
    This tutorial shows 
            1) how to download and render neurons from the MouseLight project
                    using morphapi.
            2) how to these neurons with a few different options to color them

    You can also download data manually from the neuronbrowser website and render them by
    passing the downloaded files to `scene.add_neurons`.
"""
from brainrender.scene import Scene

from morphapi.api.mouselight import MouseLightAPI


# ---------------------------- Downloading neurons --------------------------- #
mlapi = MouseLightAPI()

# Fetch metadata for neurons with some in the secondary motor cortex
neurons_metadata = mlapi.fetch_neurons_metadata(
    filterby="soma", filter_regions=["MOs"]
)

# Then we can download the files and save them as a .json file
neurons = mlapi.download_neurons(neurons_metadata[:5])


# ----------------------------- Rendering neurons ---------------------------- #

# 1 color all neurons of the same color, don't show axons
scene = Scene(title="One color")
scene.add_neurons(
    neurons, color="salmon", display_axon=False, neurite_radius=6
)
scene.render()
scene.close()

# 2 color each neuron of a different color using a colormap
scene = Scene(title="Colormap")
scene.add_neurons(neurons, color="Reds", alpha=0.8)
scene.render()
scene.close()

# 3 specify a color for each neuron
scene = Scene(title="Color each")
scene.add_neurons(
    neurons,
    color=["salmon", "darkseagreeb", "skyblue", "chocolate", "darkgoldenrod"],
    alpha=0.8,
)
scene.render()
scene.close()

# 4 specify a color for each neuronal component
scene = Scene(title="Color components")
scene.add_neurons(
    neurons, color=dict(soma="red", dendrites="orange", axon="blackboard")
)
scene.render()
scene.close()


""" 
    For more options check the add_neurons function docstring (atlases.aba.Aba -> add_neurons).
    Also check 'colored_neurons.py' example for using a custom colormap.
"""



import brainrender

brainrender.SHADER_STYLE = "cartoon"


from brainrender.scene import Scene
from brainrender.atlases.custom_atlaseselegans import Celegans

# Specify path to folder with c.elegans connectome

"""
        Currently Celegans assumes that you already possess the data.
        If this is not the case, please do get in touch. 
        We will be adding support to automatically download these data soon!
"""

data_folder = "/Users/federicoclaudi/Dropbox (UCL - SWC)/Rotation_vte/Anatomy/Atlases/atlasesforbrainrender/CELEGANS"


# --------------------------- Scene 1: all neurons --------------------------- #

# Create a brainrender scene with a custom Atlas class
"""
    By passing a custom Atlas class (instance of brainrender.atlases.base Atlas class)
    to Scene, Scene will use the atlas class' methods to fetch data and crate actors.
"""
scene = Scene(
    add_root=False,
    atlas=Celegans,  # Pass the custom atlas class to scene.
    display_inset=False,
    atlas_kwargs=dict(
        data_folder=data_folder
    ),  # use this to pass keyword arguments to the Atlas class
    title="Whole connectome",
)


# Exclude some neurons we don't want to render
metadata = scene.atlas.neurons_metadata
neurons = metadata.loc[
    (metadata.type != "nonvalid") & (metadata.type != "other")
]


# Add each neuron with the corresponding outline
scene.add_neurons(list(neurons.neuron.values))
for neuron in scene.actors["neurons"]:
    scene.add_actor(neuron.silhouette().lw(3).c("k"))

# Render
scene.render()
scene.close()

# ------------------------- Scene 2: showing synapses ------------------------ #


scene = Scene(
    add_root=False,
    atlas=Celegans,  # Pass the custom atlas class to scene.
    display_inset=False,
    atlas_kwargs=dict(
        data_folder=data_folder
    ),  # use this to pass keyword arguments to the Atlas class
    title="Synapses",
)

# Show only a few neurons and their pre and post synapses
scene.add_neurons(
    list(neurons.neuron.values[:5]), as_skeleton=True
)  # You can render the neruons as just their skeleton

scene.add_neurons_synapses(
    list(neurons.neuron.values[:5]), pre=True, post=True
)
"""
        Pre-synaptic locations are shown as grey spheres, 
        post-synaptic ones as arrows point towards the post-synaptic target.
"""

scene.render()



"""
This example shows how to crate an animate scene where
over a number of frames:
    - Different sets of neurons from the mouselight database are shown
    - The camera moves around 
"""


import brainrender

brainrender.WHOLE_SCREEN = False
brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 0.1


from brainrender.scene import Scene
import time
import numpy as np
from random import choices
from rich.progress import track
from morphapi.api.mouselight import MouseLightAPI


from brainrender.Utils.camera import (
    buildcam,
    sagittal_camera,
)
from brainrender.colors import colorMap

# --------------------------------- Variables -------------------------------- #
minalpha = 0.01  # transparency of background neurons
darkcolor = "lightgray"  # background neurons color
lightcolor = "lawngreen"  # highlighted neurons color

N_FRAMES = 250
N_neurons = 50  # number of neurons to show in total, if -1 all neurons are shown but it might take a while to render them at first
N_neurons_in_frame = 8  # number of neurons to be highlighted in a given frame
N_frames_for_change = 15  # every N frames which neurons are shown changes

# Variables to specify camera position at each frame
zoom = np.linspace(1, 1.5, N_FRAMES)
frac = np.zeros_like(
    zoom
)  # for camera transition, interpolation value between cameras
frac[:150] = np.linspace(0, 1, 150)
frac[150:] = np.linspace(1, 0, len(frac[150:]))

# -------------------------------- Fetch data -------------------------------- #

# Then we can download the files and save them as a .json file
ml_api = MouseLightAPI()
# Fetch metadata for neurons with some in the secondary motor cortex
neurons_metadata = ml_api.fetch_neurons_metadata(
    filterby="soma", filter_regions=["MOs"]
)

neurons_files = ml_api.download_neurons(neurons_metadata[:N_neurons])

# ------------------------------- Create scene ------------------------------- #
scene = Scene(display_inset=False, use_default_key_bindings=True)

neurons = scene.add_neurons(neurons_files, neurite_radius=12, alpha=0)

# Make all neurons background
for neuron in neurons:
    for mesh in neuron.values():
        mesh.alpha(minalpha)
        mesh.color(darkcolor)


# Create new cameras
cam1 = buildcam(sagittal_camera)


cam2 = buildcam(
    dict(
        position=[-16624.081, -33431.408, 33527.412],
        focal=[6587.835, 3849.085, 5688.164],
        viewup=[0.634, -0.676, -0.376],
        distance=51996.653,
        clipping=[34765.671, 73812.327],
    )
)

cam3 = buildcam(
    dict(
        position=[1862.135, -4020.792, -36292.348],
        focal=[6587.835, 3849.085, 5688.164],
        viewup=[0.185, -0.97, 0.161],
        distance=42972.44,
        clipping=[29629.503, 59872.10],
    )
)

# Iniziale camera position
startcam = scene.plotter.moveCamera(cam1, cam2, frac[0])


# ------------------------------- Create frames ------------------------------ #
# Create frames
prev_neurons = []
for step in track(
    np.arange(N_FRAMES), total=N_FRAMES, description="Generating frames..."
):
    if step % N_frames_for_change == 0:  # change neurons every N framse

        # reset neurons from previous set of neurons
        for neuron in prev_neurons:
            for component, actor in neuron.items():
                actor.alpha(minalpha)
                actor.color(darkcolor)
        prev_neurons = []

        # highlight new neurons
        neurons = choices(neurons, k=N_neurons_in_frame)
        for n, neuron in enumerate(neurons):
            color = colorMap(
                n, "Greens_r", vmin=-2, vmax=N_neurons_in_frame + 3
            )
            for component, actor in neuron.items():
                actor.alpha(1)
                actor.color(color)
            prev_neurons.append(neuron)

    # Move scene camera between 3 cameras
    scene.plotter.moveCamera(cam1, cam2, frac[step])
    if frac[step] == 1:
        cam1 = cam3

    # Update rendered window
    time.sleep(0.1)
    scene.render(zoom=zoom[step], interactive=False, video=True)



"""
This example shows how to crate an animate scene where
over a number of frames:
    - Different sets of neurons from the mouselight database are shown
    - The camera moves around 

Note that creating videos requires opencv which is not installed
with brainrender by default, please install with "pip install opencv-python"

"""


import brainrender

brainrender.WHOLE_SCREEN = False
brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 0.1


from brainrender.scene import Scene
import numpy as np
from random import choices
from rich.progress import track


from brainrender.Utils.camera import (
    top_camera,
    buildcam,
    sagittal_camera,
)
from brainrender.animation.video import CustomVideoMaker


# --------------------------------- Variables -------------------------------- #
minalpha = 0  # transparency of background neurons
darkcolor = "lightgray"  # background neurons color
lightcolor = "lawngreen"  # highlighted neurons color

N_FRAMES = 250
N_streamlines = (
    -1
)  # number of streamlines to show in total, if -1 all streamlines are shown but it might take a while to render them at first
N_streamlines_in_frame = (
    2  # number of streamlines to be highlighted in a given frame
)
N_frames_for_change = 50  # every N frames which streamlines are shown changes

# Variables to specify camera position at each frame
zoom = np.linspace(1, 1.35, N_FRAMES)
frac = np.zeros_like(
    zoom
)  # for camera transition, interpolation value between cameras
frac[:150] = np.linspace(0, 1, 150)
frac[150:] = np.linspace(1, 0, len(frac[150:]))

# ------------------------------- Create scene ------------------------------- #
scene = Scene(display_inset=True, use_default_key_bindings=True)
root = scene.actors["root"]

filepaths, data = scene.atlas.download_streamlines_for_region("TH")
scene.add_streamlines(data, color="darkseagreen", show_injection_site=False)

scene.add_brain_regions(["TH"], alpha=0.2)

# Make all streamlines background
for mesh in scene.actors["tracts"]:
    mesh.alpha(minalpha)
    mesh.color(darkcolor)


# Create new cameras
cam1 = buildcam(sagittal_camera)
cam2 = buildcam(top_camera)
cam3 = buildcam(
    dict(
        position=[1862.135, -4020.792, -36292.348],
        focal=[6587.835, 3849.085, 5688.164],
        viewup=[0.185, -0.97, 0.161],
        distance=42972.44,
        clipping=[29629.503, 59872.10],
    )
)

# Iniziale camera position
startcam = scene.plotter.moveCamera(cam1, cam2, frac[0])


# ------------------------------- Create frames ------------------------------ #
def frame_maker(scene=None, video=None, videomaker=None):
    prev_streamlines = []
    for step in track(
        np.arange(N_FRAMES), total=N_FRAMES, description="Generating frames..."
    ):
        if step % N_frames_for_change == 0:  # change neurons every N framse

            # reset neurons from previous set of neurons
            for mesh in prev_streamlines:
                mesh.alpha(minalpha)
                mesh.color(darkcolor)
            prev_streamlines = []

            # highlight new neurons
            streamlines = choices(
                scene.actors["tracts"], k=N_streamlines_in_frame
            )
            for n, mesh in enumerate(streamlines):
                # color = colorMap(n, 'Reds', vmin=-2, vmax=N_streamlines_in_frame+3)
                mesh.alpha(0.7)
                mesh.color("orangered")
                prev_streamlines.append(mesh)

        # Move scene camera between 3 cameras
        if step < 150:
            scene.plotter.moveCamera(cam1, cam2, frac[step])
        else:
            scene.plotter.moveCamera(cam3, cam2, frac[step])

        # Add frame to video
        scene.render(zoom=zoom[step], interactive=False, video=True)
        video.addFrame()
    return video


# ---------------------------------------------------------------------------- #
#                                  Video maker                                 #
# ---------------------------------------------------------------------------- #
vm = CustomVideoMaker(scene, save_name="streamlines_animation")
vm.make_video(frame_maker)



from brainrender.Utils.ABA.volumetric.VolumetricConnectomeAPI import (
    VolumetricAPI,
)


class InteractiveVolumetric(VolumetricAPI):
    def __init__(
        self,
        target_region,
        *args,
        projection_type="afferent",
        cmap=None,
        **kwargs,
    ):
        if projection_type == "afferent":
            title = f"Projections to {target_region}"
        else:
            title = f"Projections from {target_region}"
        VolumetricAPI.__init__(
            self, *args, scene_kwargs=dict(title=title), **kwargs
        )
        self.cmap = cmap

        # Add brain region
        self.target_region = target_region
        self.scene.add_brain_regions([target_region], alpha=0.4)

        # Get region mesh data
        self.target_region_mesh = self.scene.get_region_unilateral(
            target_region, hemisphere="right"
        )
        self.target_region_bounds = self.target_region_mesh.bounds()

        # The position of the point
        self.target = self.scene.atlas.get_region_CenterOfMass(
            target_region
        )  # initialise with a point in the right hemisphere
        self.target_in_region = True
        self.crosshair = []
        self._move_crosshair()

        # Projection
        self.projection_type = projection_type
        self.lego = None
        self._update_projection()

        # Add sliders
        self.scene.plotter.addSlider2D(
            self.slider_ap,
            self.target_region_bounds[0],
            self.target_region_bounds[1],
            value=self.target[0],
            pos=3,
            title="AP coordinate",
        )

        self.scene.plotter.addSlider2D(
            self.slider_dv,
            self.target_region_bounds[2],
            self.target_region_bounds[3],
            value=self.target[1],
            pos=11,
            title="DV coordinate",
        )

        self.scene.plotter.addSlider2D(
            self.slider_ml,
            self.target_region_bounds[4],
            self.target_region_bounds[5],
            value=self.target[2],
            pos=12,
            title="ML coordinate",
        )

    def _reset_actors(self):
        self.scene.actors["others"] = []

    def _move_crosshair(self):
        old_cross = self.crosshair.copy()

        if self.target_in_region:
            self.crosshair = self.scene.add_crosshair_at_point(self.target)
        else:
            self.crosshair = self.scene.add_crosshair_at_point(
                self.target,
                line_kwargs=dict(alpha=0.25),
                point_kwargs=dict(alpha=0.25),
            )

        for act in self.crosshair:
            self.scene.plotter.add(act)

        for act in old_cross:
            self.scene.plotter.remove(act)

    def _update_projection(self):
        if self.lego is not None:
            self.scene.plotter.remove(self.lego)

        if self.target_in_region:
            if self.projection_type == "afferent":
                func = self.add_mapped_projection_to_point
            else:
                func = self.add_mapped_projection_from_point

            self.lego = func(
                self.target,
                vmin=0.000001,
                show_point=False,
                show_poin_region=False,
                show_crosshair=False,
                cmap=self.cmap,
            )
            self.scene.plotter.add(self.lego)
        else:
            self.lego = None

    def _update(self):
        if self.scene._check_point_in_region(
            self.target, self.target_region_mesh
        ):
            self.target_in_region = True
        else:
            self.target_in_region = False

        self._reset_actors()
        self._move_crosshair()
        self._update_projection()

    def slider_ap(self, widget, event):
        self.target[0] = widget.GetRepresentation().GetValue()
        self._update()

    def slider_dv(self, widget, event):
        self.target[1] = widget.GetRepresentation().GetValue()
        self._update()

    def slider_ml(self, widget, event):
        self.target[2] = widget.GetRepresentation().GetValue()
        self._update()



from brainrender.scene import Scene

# brainrender.SCREENSHOT_TRANSPARENT_BACKGROUND = False

scene = Scene()

scene.add_brain_regions(["CA1"])

scene.add_optic_cannula(target_region="CA1")

p0 = scene.atlas.get_region_CenterOfMass("MOs")
scene.add_optic_cannula(pos=p0)

scene.render()



import numpy as np

from vedo import merge
from vedo.shapes import Line, Sphere, Text
from vedo.utils import precision, mag


def ruler(p1, p2, unit_scale=1, units=None, s=50):
    actors = []

    # Make two line segments
    midpoint = np.array([(x + y) / 2 for x, y in zip(p1, p2)])
    gap1 = ((midpoint - p1) * 0.8) + p1
    gap2 = ((midpoint - p2) * 0.8) + p2

    actors.append(Line(p1, gap1, lw=2))
    actors.append(Line(gap2, p2, lw=2))

    # Add label
    if units is None:
        units = ""
    dist = mag(p2 - p1) * unit_scale
    label = precision(dist, 3) + " " + units
    lbl = Text(label, pos=midpoint, s=s + 100, justify="center")
    lbl.SetOrientation([0, 0, 180])
    actors.append(lbl)

    # Add spheres add end
    actors.append(Sphere(p1, r=s, c=[0.3, 0.3, 0.3]))
    actors.append(Sphere(p2, r=s, c=[0.3, 0.3, 0.3]))

    acts = merge(*actors).c((0.3, 0.3, 0.3)).alpha(1).lw(2)
    acts.name = "Ruler"
    acts._bg_class = "Ruler"
    return acts



import vtk
from accepts import accepts

sagittal_camera = dict(
    position=[5532.562, 8057.256, 66760.941],
    focal=[6587.835, 3849.085, 5688.164],
    viewup=[0.045, -0.997, 0.061],
    distance=24893.917,
    clipping=[9826.898, 43920.235],
)

sagittal_camera2 = dict(
    position=(10705.845660949382, 7435.678067378925, -36936.3695486442),
    focal=(6779.790352916297, 3916.3916231239214, 5711.389387062087),
    viewup=(-0.0050579179155257475, -0.9965615097647067, -0.08270172139591858),
    distance=42972.44034956088,
    clipping=(30461.81976236306, 58824.38622122339),
)


coronal_camera = dict(
    position=[-54546.708, 526.426, 6171.914],
    focal=[6587.835, 3849.085, 5688.164],
    viewup=[0.054, -0.999, -0.002],
    distance=61226.68,
    clipping=[47007.899, 79272.88],
)

top_camera = dict(
    position=[1482.128, -57162.314, 5190.803],
    focal=[6587.835, 3849.085, 5688.164],
    viewup=[-0.996, 0.083, 0.016],
    distance=61226.68,
    clipping=[52067.13, 72904.35],
)


three_quarters_camera = dict(
    position=[-27271.044, -18212.523, 51683.47],
    focal=[6587.835, 3849.085, 5688.164],
    viewup=[0.341, -0.921, -0.19],
    distance=61226.68,
    clipping=[42904.747, 84437.903],
)

cameras = dict(
    sagittal=sagittal_camera,
    sagittal2=sagittal_camera2,
    coronal=coronal_camera,
    top=top_camera,
    three_quarters=three_quarters_camera,
)


@accepts((str, dict, list))
def check_camera_param(camera):
    if isinstance(camera, str):
        if camera not in cameras.keys():
            raise ValueError(
                f"Camera name {camera} if not recognized.\nValid cameras: {cameras.keys()}"
            )
        cam = cameras[camera]
        if not len(cam):
            raise ValueError(f"Camera name {camera} is an empty dictionary!")
        return cam

    elif not isinstance(camera, dict):
        raise ValueError(
            "Camera should be either a string with a camera name or a dictionary of camera params."
        )

    else:
        params = ["position", "focal", "viewup", "distance", "clipping"]
        for param in params:
            if param not in list(camera.keys()):
                raise ValueError(
                    f"Camera parameters dict should include the following keys: {params}"
                )
        return camera


def buildcam(cm):
    """
        Builds a camera from a dictionary of parameters, from vedo
    """

    cm = cm.copy()

    cm_pos = cm.pop("position", None)
    cm_focalPoint = cm.pop("focal", None)
    cm_viewup = cm.pop("viewup", None)
    cm_distance = cm.pop("distance", None)
    cm_clippingRange = cm.pop("clipping", None)
    cm_parallelScale = cm.pop("parallelScale", None)
    cm_thickness = cm.pop("thickness", None)
    cm_viewAngle = cm.pop("viewAngle", None)

    cm = vtk.vtkCamera()

    if cm_pos is not None:
        cm.SetPosition(cm_pos)
    if cm_focalPoint is not None:
        cm.SetFocalPoint(cm_focalPoint)
    if cm_viewup is not None:
        cm.SetViewUp(cm_viewup)
    if cm_distance is not None:
        cm.SetDistance(cm_distance)
    if cm_clippingRange is not None:
        cm.SetClippingRange(cm_clippingRange)
    if cm_parallelScale is not None:
        cm.SetParallelScale(cm_parallelScale)
    if cm_thickness is not None:
        cm.SetThickness(cm_thickness)
    if cm_viewAngle is not None:
        cm.SetViewAngle(cm_viewAngle)

    return cm


def set_camera_params(camera, params):
    # Apply camera parameters
    camera.SetPosition(params["position"])
    camera.SetFocalPoint(params["focal"])
    camera.SetViewUp(params["viewup"])
    camera.SetDistance(params["distance"])
    camera.SetClippingRange(params["clipping"])


def set_camera(scene, camera):
    """
        Sets the position of the camera of a brainrender scene.

        :param scene: instance of Scene()
        :param camera: either a string with the name of one of the available cameras, or
                        a dictionary of camera parameters. 
        
    """
    if camera is None:
        return

    if not isinstance(camera, vtk.vtkCamera):
        # Get camera params
        camera = check_camera_param(camera)

        # set params
        set_camera_params(scene.plotter.camera, camera)
    else:
        scene.plotter.camera = camera
    return camera


def get_camera_params(scene=None, camera=None):
    """
        Given an active brainrender scene, it return
        the camera parameters. 
    """
    if scene is not None:
        if not scene.is_rendered:
            scene.render(interactive=False)
        cam = scene.plotter.camera
    else:
        cam = camera

    params = dict(
        position=cam.GetPosition(),
        focal=cam.GetFocalPoint(),
        viewup=cam.GetViewUp(),
        distance=cam.GetDistance(),
        clipping=cam.GetClippingRange(),
        orientation=cam.GetOrientation(),
    )
    return params



import numpy as np
import random
import vtk

# The following code and colors list is from vedo.color : https://github.com/marcomusy/vedo/blob/master/vedo/colors.py
# The code is copied here just to make it easier to look up and change colros
try:
    import matplotlib
    import matplotlib.cm as cm_mpl

    _mapscales = cm_mpl
except:
    _mapscales = None
    # see below, this is dealt with in colorMap()


#########################################################
# basic color schemes
#########################################################
# from matplotlib
colors = {
    "aliceblue": "#F0F8FF",
    "antiquewhite": "#FAEBD7",
    "aqua": "#00FFFF",
    "aquamarine": "#7FFFD4",
    "azure": "#F0FFFF",
    "beige": "#F5F5DC",
    "bisque": "#FFE4C4",
    "blanchedalmond": "#FFEBCD",
    "blue": "#0000FF",
    "blueviolet": "#8A2BE2",
    "brown": "#A52A2A",
    "burlywood": "#DEB887",
    "cadetblue": "#5F9EA0",
    "chartreuse": "#7FFF00",
    "chocolate": "#D2691E",
    "coral": "#FF7F50",
    "cornflowerblue": "#6495ED",
    "cornsilk": "#FFF8DC",
    "crimson": "#DC143C",
    "cyan": "#00FFFF",
    "darkblue": "#00008B",
    "darkcyan": "#008B8B",
    "darkgoldenrod": "#B8860B",
    "darkgray": "#A9A9A9",
    "darkgreen": "#006400",
    "darkkhaki": "#BDB76B",
    "darkmagenta": "#8B008B",
    "darkolivegreen": "#556B2F",
    "darkorange": "#FF8C00",
    "darkorchid": "#9932CC",
    "darkred": "#8B0000",
    "darksalmon": "#E9967A",
    "darkseagreen": "#8FBC8F",
    "darkslateblue": "#483D8B",
    "darkslategray": "#2F4F4F",
    "darkturquoise": "#00CED1",
    "darkviolet": "#9400D3",
    "deeppink": "#FF1493",
    "deepskyblue": "#00BFFF",
    "dimgray": "#696969",
    "dodgerblue": "#1E90FF",
    "firebrick": "#B22222",
    "floralwhite": "#FFFAF0",
    "forestgreen": "#228B22",
    "fuchsia": "#FF00FF",
    "gainsboro": "#DCDCDC",
    "ghostwhite": "#F8F8FF",
    "gold": "#FFD700",
    "goldenrod": "#DAA520",
    "gray": "#808080",
    "green": "#008000",
    "greenyellow": "#ADFF2F",
    "honeydew": "#F0FFF0",
    "hotpink": "#FF69B4",
    "indianred": "#CD5C5C",
    "indigo": "#4B0082",
    "ivory": "#FFFFF0",
    "khaki": "#F0E68C",
    "lavender": "#E6E6FA",
    "lavenderblush": "#FFF0F5",
    "lawngreen": "#7CFC00",
    "lemonchiffon": "#FFFACD",
    "lightblue": "#ADD8E6",
    "lightcoral": "#F08080",
    "lightcyan": "#E0FFFF",
    "lightgray": "#D3D3D3",
    "lightgreen": "#90EE90",
    "lightpink": "#FFB6C1",
    "lightsalmon": "#FFA07A",
    "lightseagreen": "#20B2AA",
    "lightskyblue": "#87CEFA",
    "lightsteelblue": "#B0C4DE",
    "lightyellow": "#FFFFE0",
    "lime": "#00FF00",
    "limegreen": "#32CD32",
    "linen": "#FAF0E6",
    "magenta": "#FF00FF",
    "maroon": "#800000",
    "mediumaquamarine": "#66CDAA",
    "mediumblue": "#0000CD",
    "mediumorchid": "#BA55D3",
    "mediumpurple": "#9370DB",
    "mediumseagreen": "#3CB371",
    "mediumslateblue": "#7B68EE",
    "mediumspringgreen": "#00FA9A",
    "mediumturquoise": "#48D1CC",
    "mediumvioletred": "#C71585",
    "midnightblue": "#191970",
    "mintcream": "#F5FFFA",
    "mistyrose": "#FFE4E1",
    "moccasin": "#FFE4B5",
    "navajowhite": "#FFDEAD",
    "navy": "#000080",
    "oldlace": "#FDF5E6",
    "olive": "#808000",
    "olivedrab": "#6B8E23",
    "orange": "#FFA500",
    "orangered": "#FF4500",
    "orchid": "#DA70D6",
    "palegoldenrod": "#EEE8AA",
    "palegreen": "#98FB98",
    "paleturquoise": "#AFEEEE",
    "palevioletred": "#DB7093",
    "papayawhip": "#FFEFD5",
    "peachpuff": "#FFDAB9",
    "peru": "#CD853F",
    "pink": "#FFC0CB",
    "plum": "#DDA0DD",
    "powderblue": "#B0E0E6",
    "purple": "#800080",
    "rebeccapurple": "#663399",
    "red": "#FF0000",
    "rosybrown": "#BC8F8F",
    "royalblue": "#4169E1",
    "saddlebrown": "#8B4513",
    "salmon": "#FA8072",
    "sandybrown": "#F4A460",
    "seagreen": "#2E8B57",
    "seashell": "#FFF5EE",
    "sienna": "#A0522D",
    "silver": "#C0C0C0",
    "skyblue": "#87CEEB",
    "slateblue": "#6A5ACD",
    "slategray": "#708090",
    "snow": "#FFFAFA",
    "blackboard": "#393939",
    "springgreen": "#00FF7F",
    "steelblue": "#4682B4",
    "tan": "#D2B48C",
    "teal": "#008080",
    "thistle": "#D8BFD8",
    "tomato": "#FF6347",
    "turquoise": "#40E0D0",
    "violet": "#EE82EE",
    "wheat": "#F5DEB3",
    "white": "#FFFFFF",
    "whitesmoke": "#F5F5F5",
    "yellow": "#FFFF00",
    "yellowgreen": "#9ACD32",
}

# color nicknames
color_nicks = {
    "a": "aqua",
    "b": "blue",
    "bb": "blackboard",
    "c": "cyan",
    "f": "fuchsia",
    "g": "green",
    "i": "indigo",
    "m": "magenta",
    "n": "navy",
    "l": "lavender",
    "o": "orange",
    "p": "purple",
    "r": "red",
    "s": "salmon",
    "t": "tomato",
    "v": "violet",
    "y": "yellow",
    "w": "white",
    "lb": "lightblue",  # light
    "lg": "lightgreen",
    "lr": "orangered",
    "lc": "lightcyan",
    "ls": "lightsalmon",
    "ly": "lightyellow",
    "dr": "darkred",  # dark
    "db": "darkblue",
    "dg": "darkgreen",
    "dm": "darkmagenta",
    "dc": "darkcyan",
    "ds": "darksalmon",
    "dv": "darkviolet",
}

# available colormap names from matplotlib:
_mapscales_cmaps = (
    "Accent",
    "Accent_r",
    "Blues",
    "Blues_r",
    "BrBG",
    "BrBG_r",
    "BuGn",
    "BuGn_r",
    "BuPu",
    "BuPu_r",
    "CMRmap",
    "CMRmap_r",
    "Dark2",
    "Dark2_r",
    "GnBu",
    "GnBu_r",
    "Greens",
    "Greens_r",
    "Greys",
    "Greys_r",
    "OrRd",
    "OrRd_r",
    "Oranges",
    "Oranges_r",
    "PRGn",
    "PRGn_r",
    "Paired",
    "Paired_r",
    "Pastel1",
    "Pastel1_r",
    "Pastel2",
    "Pastel2_r",
    "PiYG",
    "PiYG_r",
    "PuBu",
    "PuBuGn",
    "PuBuGn_r",
    "PuBu_r",
    "PuOr",
    "PuOr_r",
    "PuRd",
    "PuRd_r",
    "Purples",
    "Purples_r",
    "RdBu",
    "RdBu_r",
    "RdGy",
    "RdGy_r",
    "RdPu",
    "RdPu_r",
    "RdYlBu",
    "RdYlBu_r",
    "RdYlGn",
    "RdYlGn_r",
    "Reds",
    "Reds_r",
    "Set1",
    "Set1_r",
    "Set2",
    "Set2_r",
    "Set3",
    "Set3_r",
    "Spectral",
    "Spectral_r",
    "Wistia",
    "Wistia_r",
    "YlGn",
    "YlGnBu",
    "YlGnBu_r",
    "YlGn_r",
    "YlOrBr",
    "YlOrBr_r",
    "YlOrRd",
    "YlOrRd_r",
    "afmhot",
    "afmhot_r",
    "autumn",
    "autumn_r",
    "binary",
    "binary_r",
    "bone",
    "bone_r",
    "brg",
    "brg_r",
    "bwr",
    "bwr_r",
    "cividis",
    "cividis_r",
    "cool",
    "cool_r",
    "coolwarm",
    "coolwarm_r",
    "copper",
    "copper_r",
    "cubehelix",
    "cubehelix_r",
    "flag",
    "flag_r",
    "gist_earth",
    "gist_earth_r",
    "gist_gray",
    "gist_gray_r",
    "gist_heat",
    "gist_heat_r",
    "gist_ncar",
    "gist_ncar_r",
    "gist_rainbow",
    "gist_rainbow_r",
    "gist_stern",
    "gist_stern_r",
    "gist_yarg",
    "gist_yarg_r",
    "gnuplot",
    "gnuplot2",
    "gnuplot2_r",
    "gnuplot_r",
    "gray_r",
    "hot",
    "hot_r",
    "hsv",
    "hsv_r",
    "inferno",
    "inferno_r",
    "jet",
    "jet_r",
    "magma",
    "magma_r",
    "nipy_spectral",
    "nipy_spectral_r",
    "ocean",
    "ocean_r",
    "pink_r",
    "plasma",
    "plasma_r",
    "prism",
    "prism_r",
    "rainbow",
    "rainbow_r",
    "seismic",
    "seismic_r",
    "spring",
    "spring_r",
    "summer",
    "summer_r",
    "tab10",
    "tab10_r",
    "tab20",
    "tab20_r",
    "tab20b",
    "tab20b_r",
    "tab20c",
    "tab20c_r",
    "terrain",
    "terrain_r",
    "twilight",
    "twilight_r",
    "twilight_shifted",
    "twilight_shifted_r",
    "viridis",
    "viridis_r",
    "winter",
    "winter_r",
)

# default sets of colors
colors1 = [
    [1.0, 0.832, 0.000],  # gold
    [0.960, 0.509, 0.188],
    [0.901, 0.098, 0.194],
    [0.235, 0.85, 0.294],
    [0.46, 0.48, 0.000],
    [0.274, 0.941, 0.941],
    [0.0, 0.509, 0.784],
    [0.1, 0.1, 0.900],
    [0.902, 0.7, 1.000],
    [0.941, 0.196, 0.901],
]
# negative integer color number get this:
colors2 = [
    (0.99, 0.83, 0),  # gold
    (0.59, 0.0, 0.09),  # dark red
    (0.5, 1.0, 0.0),  # green
    (0.5, 0.5, 0),  # yellow-green
    (0.0, 0.66, 0.42),  # green blue
    (0.0, 0.18, 0.65),  # blue
    (0.4, 0.0, 0.4),  # plum
    (0.4, 0.0, 0.6),
    (0.2, 0.4, 0.6),
    (0.1, 0.3, 0.2),
]


def get_random_colormap():
    return random.choice(_mapscales_cmaps)


def get_n_shades_of(shade, n):
    """
    :param shade:  color
    :param n: numnber of colors

    """
    shades = [k for k, v in colors.items() if shade in k]
    if not shades:
        raise ValueError("Could not find shades for {}".format(shade))
    else:
        return random.choices(shades, k=n)


def _isSequence(arg):
    """

    :param arg:   item to check

    """
    # Check if input is iterable.
    if hasattr(arg, "strip"):
        return False
    if hasattr(arg, "__getslice__"):
        return True
    if hasattr(arg, "__iter__"):
        return True
    return False


def getColor(rgb=None, hsv=None):
    """Convert a color or list of colors to (r,g,b) format from many different input formats.

    :param bool: hsv: if set to `True`, rgb is assumed as (hue, saturation, value).
    Example:
         - RGB    = (255, 255, 255), corresponds to white
         - rgb    = (1,1,1) is white
         - hex    = #FFFF00 is yellow
         - string = 'white'
         - string = 'w' is white nickname
         - string = 'dr' is darkred
         - int    =  7 picks color nr. 7 in a predefined color list
         - int    = -7 picks color nr. 7 in a different predefined list
    |colorcubes| |colorcubes.py|_
    :param rgb:  (Default value = None)
    :param hsv:  (Default value = None)

    """
    # recursion, return a list if input is list of colors:
    if _isSequence(rgb) and (len(rgb) > 3 or _isSequence(rgb[0])):
        seqcol = []
        for sc in rgb:
            seqcol.append(getColor(sc))
        return seqcol

    if str(rgb).isdigit():
        rgb = int(rgb)

    if hsv:
        c = hsv2rgb(hsv)
    else:
        c = rgb

    if _isSequence(c):
        if c[0] <= 1 and c[1] <= 1 and c[2] <= 1:
            return c  # already rgb
        else:
            if len(c) == 3:
                return list(np.array(c) / 255.0)  # RGB
            else:
                return (c[0] / 255.0, c[1] / 255.0, c[2] / 255.0, c[3])  # RGBA

    elif isinstance(c, str):  # is string
        c = c.replace("grey", "gray").replace(" ", "")
        if 0 < len(c) < 3:  # single/double letter color
            if c.lower() in color_nicks.keys():
                c = color_nicks[c.lower()]
            else:
                print("Unknow color nickname:", c)
                print("Available abbreviations:", color_nicks)
                return (0.5, 0.5, 0.5)

        if c.lower() in colors.keys():  # matplotlib name color
            c = colors[c.lower()]
        else:  # vtk name color
            namedColors = vtk.vtkNamedColors()
            rgba = [0, 0, 0, 0]
            namedColors.GetColor(c, rgba)
            return list(np.array(rgba[0:3]) / 255.0)

        if "#" in c:  # hex to rgb
            h = c.lstrip("#")
            rgb255 = list(int(h[i : i + 2], 16) for i in (0, 2, 4))
            rgbh = np.array(rgb255) / 255.0
            if np.sum(rgbh) > 3:
                print("Error in getColor(): Wrong hex color", c)
                return (0.5, 0.5, 0.5)
            return tuple(rgbh)

    elif isinstance(c, int):  # color number
        if c >= 0:
            return colors1[c % 10]
        else:
            return colors2[-c % 10]

    elif isinstance(c, float):
        if c >= 0:
            return colors1[int(c) % 10]
        else:
            return colors2[int(-c) % 10]

    # print("Unknown color:", c)
    return (0.5, 0.5, 0.5)


def getColorName(c):
    """Find the name of a color.
    |colorpalette| |colorpalette.py|_

    :param c: c 

    """
    c = np.array(getColor(c))  # reformat to rgb
    mdist = 99.0
    kclosest = ""
    for key in colors.keys():
        ci = np.array(getColor(key))
        d = np.linalg.norm(c - ci)
        if d < mdist:
            mdist = d
            kclosest = str(key)
    return kclosest


def hsv2rgb(hsv):
    """Convert HSV to RGB color.

    :param hsv: gsv 

    """
    ma = vtk.vtkMath()
    return ma.HSVToRGB(hsv)


def rgb2hsv(rgb):
    """Convert RGB to HSV color.

    :param rgb: rgb 

    """
    ma = vtk.vtkMath()
    return ma.RGBToHSV(getColor(rgb))


def rgb2int(rgb_tuple):
    """

    :param rgb_tuple: rgb 

    """
    rgb = (
        int(rgb_tuple[0] * 255),
        int(rgb_tuple[1] * 255),
        int(rgb_tuple[2] * 255),
    )
    return 65536 * rgb[0] + 256 * rgb[1] + rgb[2]


def colorMap(value, name="jet", vmin=None, vmax=None):
    """Map a real value in range [vmin, vmax] to a (r,g,b) color scale.

    :param value: scalar value to transform into a color
    :type value: float, list
    :param name: color map name (Default value = "jet")
    :type name: str, matplotlib.colors.LinearSegmentedColormap
    :param vmin:  (Default value = None)
    :param vmax:  (Default value = None)
    :returns: return: (r,g,b) color, or a list of (r,g,b) colors.
    .. note:: Most frequently used color maps:
        |colormaps|
        Matplotlib full list:
        .. image:: https://matplotlib.org/1.2.1/_images/show_colormaps.png
    .. tip:: Can also use directly a matplotlib color map:
        :Example:
            .. code-block:: python
                from vedo import colorMap
                import matplotlib.cm as cm
                print( colorMap(0.2, cm.flag, 0, 1) )
                (1.0, 0.809016994374948, 0.6173258487801733)

    """
    if not _mapscales:
        print(
            "-------------------------------------------------------------------"
        )
        print(
            "WARNING : cannot import matplotlib.cm (colormaps will show up gray)."
        )
        print("Try e.g.: sudo apt-get install python3-matplotlib")
        print("     or : pip install matplotlib")
        print(
            "     or : build your own map (see example in basic/mesh_custom.py)."
        )
        return (0.5, 0.5, 0.5)

    if isinstance(name, matplotlib.colors.LinearSegmentedColormap):
        mp = name
    else:
        mp = cm_mpl.get_cmap(name=name)

    if _isSequence(value):
        values = np.array(value)
        if vmin is None:
            vmin = np.min(values)
        if vmax is None:
            vmax = np.max(values)
        values = np.clip(values, vmin, vmax)
        values -= vmin
        values = values / (vmax - vmin)
        cols = []
        mp = cm_mpl.get_cmap(name=name)
        for v in values:
            cols.append(mp(v)[0:3])
        return np.array(cols)
    else:
        value -= vmin
        value /= vmax - vmin
        if value > 0.999:
            value = 0.999
        elif value < 0:
            value = 0
        return mp(value)[0:3]


def makePalette(N, *colors):
    """Generate N colors starting from `color1` to `color2`
    by linear interpolation HSV in or RGB spaces.
    Adapted from vedo makePalette function

    :param int: N: number of output colors.
    :param colors: input colors, any number of colors with 0 < ncolors <= N is okay.
    """
    if not isinstance(N, (float, int)):
        raise ValueError(
            f"The first argument N should be an integer, not {type(N)}."
        )
    N = int(N)

    N_input_colors = len(colors)
    if not N_input_colors:
        raise ValueError("No colors where passed to makePalette")
    if N_input_colors > N:
        raise ValueError(
            "More input colors than out colors (N) where passed to makePalette"
        )

    if N_input_colors == 1:
        return [np.array(getColor(colors[0])) for n in np.arange(N)]
    if N_input_colors == N:
        return colors
    else:
        # Get how many colors for each pair of colors we are interpolating over
        fractions = [
            N // N_input_colors + (1 if x < N % N_input_colors else 0)
            for x in range(N_input_colors)
        ]

        # Get pairs of colors
        cs = [np.array(getColor(col)) for col in colors]
        cs += [cs[-1]]

        output = []
        for n, (c1, c2) in enumerate(zip(cs, cs[1:])):
            cols = []
            for f in np.linspace(0, 1, fractions[n], endpoint=True):
                c = c1 * (1 - f) + c2 * f
                cols.append(c)
            output.extend(cols)

        if len(output) != N:
            if len(output) > N:
                return output[:N]
            else:
                raise ValueError(
                    f"Expected number of output colors was {N} but we got {len(output)} instead"
                )
        return output


def get_random_colors(n_colors=1):
    """

    :param n_colors:  (Default value = 1)

    """
    if not isinstance(n_colors, np.int):
        raise ValueError("n_colors should be an integer")
    if n_colors <= 0:
        raise ValueError("n_colors should be bigger or equal to 0")

    if n_colors == 1:
        return random.choice(list(colors.keys()))
    else:
        return list(random.choices(list(colors.keys()), k=n_colors))


def check_colors(color):
    """

    :param color: color 

    """
    if isinstance(color, list):
        for col in color:
            try:
                getColor(col)
            except:
                return False
    else:
        try:
            getColor(color)
        except:
            return False
    return True



"""
    This example shows how to download and visualise volumetric
    gene expression data from the Allen Atlas
    dataset: https://mouse.brain-map.org/
"""
from brainrender.scene import Scene
from brainrender.ABA.gene_expression import GeneExpressionAPI


# ---------------------------- Download gene data ---------------------------- #
gene = "Cacna2d1"

geapi = GeneExpressionAPI()  # <- used to download the data from the allen API

"""
    To download a gene's data you need two things:
        - the id of the gene in the allen database
        - the id(s) of the ISH experiments for that gene

    you can get both with GeneExpressionAPI
"""
# Get gene id
geneid = geapi.get_gene_id_by_name(gene)  # 12078

# Get experiment IDs
expids = geapi.get_gene_experiments(
    geneid
)  # [75042246, 72119649, 74000600, 69236915]


"""
    You can then download the data for one of the experiments above
"""
data = geapi.get_gene_data(geneid, expids[0])

"""
    And finally you can take the volumetric data and turn it 
    into an actor that can be added to your brainrender scene.

    When creating the mesh it's useful to set a threshold to eliminate
    voxels with low gene expression energy. This can be done in two ways:
        - [min_value] a user defined threshold value
        - [min_quantile] a percentile (range 0-100): only voxels with value above
                the percentile are rendered.

    You can also pass any matplotlib (or custom) colormap
    to cmap to specify how the voxels will be colored
"""
gene_actor = geapi.griddata_to_volume(data, min_quantile=90, cmap="inferno")


# ---------------------------------- Render ---------------------------------- #

scene = Scene(title=gene)
scene.add_actor(gene_actor)
scene.render()



""" 
    This tutorial shows how download images for the allen institute
    mouse brain atlas. 
"""

from brainrender.ABA.atlas_images import ImageDownload

imd = ImageDownload()

# Download coronal atlas as png
fld = "atlas_images"  # pass a path to the folder where you want to save the images
imd.download_images_by_atlasid(
    fld, imd.mouse_coronal_atlas_id, atlas_svg=False
)



from setuptools import setup, find_namespace_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy",
    "pandas",
    "vedo>=2020.4.1 ",
    "k3d==2.7.4",
    "msgpack",
    "pyyaml>=5.3",
    "brainio>=0.0.19",
    "morphapi>=0.1.1.8",
    "requests",
    "bg-atlasapi>=1.0.0",
    "tables",
    "accepts",
    "pyinspect>=0.0.8",
]

setup(
    name="brainrender",
    version="1.0.1.1",
    description="Python scripts to use Allen Brain Map data for analysis "
    "and rendering",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require={
        "nb": ["jupyter", "k3d"],
        "dev": [
            "pytest-cov",
            "pytest",
            "pytest-sugar",
            "coveralls",
            "coverage<=4.5.4",
            "pre-commit",
            "opencv-python",
            "jupyter",
            "allensdk",
        ],
    },
    python_requires=">=3.6",
    packages=find_namespace_packages(
        exclude=("Installation", "Meshes", "Metadata", "Screenshots")
    ),
    include_package_data=True,
    url="https://github.com/brainglobe/brainrender",
    author="Federico Claudi",
    zip_safe=False,
    entry_points={"console_scripts": ["brainrender = brainrender.cli:main"]},
)



import pytest
from vedo import Mesh
import numpy as np

from brainrender.scene import Scene
from brainrender.colors import makePalette
from morphapi.api.mouselight import MouseLightAPI
from brainrender.ABA.gene_expression import GeneExpressionAPI


@pytest.fixture
def scene():
    return Scene()


def test_gene_expression():
    geapi = GeneExpressionAPI()
    gene = "Cacna2d1"
    geneid = geapi.get_gene_id_by_name(gene)
    symbol = geapi.get_gene_symbol_by_id(geneid)

    expids = geapi.get_gene_experiments(geneid)
    if not isinstance(geneid, int):
        raise ValueError
    if symbol != gene:
        raise ValueError
    if not isinstance(expids, list) or not len(expids):
        raise ValueError
    if not isinstance(expids[0], int):
        raise ValueError

    data = geapi.get_gene_data(geneid, expids[0])
    data3 = geapi.get_gene_data(
        geneid, expids[0]
    )  # again to make sure cache works
    data2 = geapi.get_gene_data(geneid, expids[1])

    if (
        not isinstance(data, np.ndarray)
        or not isinstance(data2, np.ndarray)
        or not isinstance(data3, np.ndarray)
    ):
        raise ValueError
    if not np.array_equal(data, data3):
        raise ValueError

    gene_actor = geapi.griddata_to_volume(
        data, min_quantile=90, cmap="inferno"
    )
    gene_actor2 = geapi.griddata_to_volume(data2, min_value=0.2)

    if not isinstance(gene_actor, Mesh) or not isinstance(gene_actor2, Mesh):
        raise ValueError


def test_streamlines(scene):
    filepaths, data = scene.atlas.download_streamlines_for_region("CA1")
    scene.add_streamlines(
        data[0], color="darkseagreen", show_injection_site=False
    )


def test_streamlines_download(scene):
    p0 = scene.atlas.get_region_CenterOfMass("ZI")
    scene.atlas.download_streamlines_to_region(p0)


def test_streamlines_colored(scene):

    filepaths, data = scene.atlas.download_streamlines_for_region("CA1")
    colors = makePalette(len(data), "salmon", "lightgreen")
    scene.add_streamlines(data, color=colors, show_injection_site=False)
    scene.add_streamlines(filepaths, color=None, show_injection_site=True)


def test_neurons(scene):
    mlapi = MouseLightAPI()
    neurons_metadata = mlapi.fetch_neurons_metadata(
        filterby="soma", filter_regions=["MOs"]
    )

    neurons = mlapi.download_neurons(neurons_metadata[:2])
    actors = scene.add_neurons(
        neurons,
        color="salmon",
        display_axon=False,
        neurite_radius=6,
        soma_radius=20,
    )

    actor = scene.add_neurons(
        neurons[0], color="salmon", display_axon=False, use_cache=True,
    )
    scene.add_neurons(
        actors, color=None, display_dendrites=False, use_cache=False
    )

    scene.add_neurons(actors, color="Reds", display_dendrites=False)

    scene.add_neurons(
        actors, color=["green" for i in actors], display_dendrites=False
    )

    scene.add_neurons(
        actors, color=dict(soma="red", axon="green", dendrites="blue")
    )

    scene.add_neurons(
        actors,
        color=[
            dict(soma="red", axon="green", dendrites="blue") for i in actors
        ],
    )

    scene.add_neurons(actor)


def test_tractography(scene):
    p0 = scene.atlas.get_region_CenterOfMass("ZI")

    tract = scene.atlas.get_projection_tracts_to_target(p0=p0)

    scene.add_tractography(tract[:25], color_by="manual", color="red")

    scene.add_tractography(
        tract[25:50],
        color_by="target_region",
        VIP_regions=["SCm"],
        VIP_color="green",
    )

    scene.add_tractography(
        tract[50:75],
        color_by="target_region",
        VIP_regions=["SCm"],
        VIP_color="green",
        include_all_inj_regions=True,
        display_injection_volume=True,
    )



"""
    This example shows how to add 3D text to your scenes
"""

import brainrender

brainrender.SHADER_STYLE = "cartoon"

from brainrender.scene import Scene
from brainrender.colors import makePalette
from vedo import Text


camera = dict(
    {
        "position": (4.58173814842751, 1.3170832519052862, 28.807539941401348),
        "focal": (5.63320729136467, 0.0, 0.13752250373363495),
        "viewup": (
            0.008857944291024357,
            0.9989221022772287,
            -0.04556501294830602,
        ),
        "distance": 28.719508970857582,
        "clipping": (27.608253763329618, 30.187623512524752),
        "orientation": (
            2.62851947276041,
            2.1003738639279446,
            0.41150325521778314,
        ),
    }
)

# Crate a scene
scene = Scene(
    add_root=False,
    display_inset=False,
    use_default_key_bindings=False,
    camera=camera,
)


# Text to add
s = "BRAINRENDER"

# Specify a color for each letter
colors = makePalette(len(s), "salmon", "powderblue")

x = 0  # use to specify the position of each letter

# Add letters one at the time to color them individually
for n, letter in enumerate("BRAINRENDER"):
    if "I" == letter or "N" == letter and n < 5:  # make the spacing right
        x += 0.6
    else:
        x += 1

    # Add letter and silhouette to the scene
    act = Text(
        letter,
        depth=0.5,
        c=colors[n],
        pos=(x, 0, 0),
        justify="centered",
        alpha=1,
    )

    scene.add_silhouette(act, lw=3)


scene.render()



import brainrender

brainrender.SHADER_STYLE = "cartoon"
brainrender.ROOT_ALPHA = 0.2
from brainrender.scene import Scene

from brainrender.atlases.custom_atlases.insects_brains_db import (
    IBDB,
)  # ! this class is the atlas for insects brains

"""
    In this tutorial you can see how to use brainrender to fetch
    and render 3D rendering of brain regions from a number of insects brains.

    All data is hosted by: https://insectbraindb.org/app/
    Please read their terms of use (https://insectbraindb.org/app/terms)
    Also please refer to the website to find the original source of these datasets

    You can also visualize these data interactively at: https://insectbraindb.org/app/three;controls=true;mode=species

"""

# Create a brainrender scene with a custom Atlas class
"""
    By passing a custom Atlas class (instance of brainrender.atlases.base Atlas class)
    to Scene, Scene will use the atlas class' methods to fetch data and crate actors.
"""
scene = Scene(
    atlas=IBDB,  # specify that we are using the insects brains databse atlas
    atlas_kwargs=dict(
        species="Schistocerca gregaria"
    ),  # Specify which insect species' brain to use
)


# Add some brain regions in the mushroom body to the rendering
central_complex = [
    "CBU-S2",
    "CBU-S1",
    "CBU-S3",
    "NO-S3_left",
    "NO-S2_left",
    "NO-S2_right",
    "NO-S3_right",
    "NO_S1_left",
    "NO-S1_right",
    "NO-S4_left",
    "NO-S4_right",
    "CBL",
    "PB",
]

scene.add_brain_regions(central_complex, alpha=1)

scene.render()



from pathlib import Path
import pandas as pd
import numpy as np

from rich.progress import track
from vedo import shapes, merge, Cylinder

import brainrender
from brainrender.Utils.data_io import (
    load_json,
    get_probe_points_from_sharptrack,
)
from brainrender.Utils.data_manipulation import is_any_item_in_list
from brainrender.Utils.webqueries import request
from brainrender.colors import check_colors

""" 
    Code to support atlases.mouse.ABA
"""


# ---------------------------------------------------------------------------- #
#                                 TRACTOGRAPHY                                 #
# ---------------------------------------------------------------------------- #
def parse_tractography_colors(
    tractography,
    include_all_inj_regions,
    color=None,
    color_by="manual",
    VIP_regions=[],
    VIP_color=None,
    others_color="salmon",
):
    """
        parses color arguments to render tracrography data

        :param tractography: list of dictionaries with tractography data
        :param color: color of rendered tractography data
        :param color_by: str, specifies which criteria to use to color the tractography (Default value = "manual")
            Options:
                - manual: the user needs to provide a color or list of colors
                - target_region: tracts are colored according to the region where the injection was done.
                        if VIP_regions is passed, then only tracts for the VIP regions are colored

        :param VIP_regions: list of brain regions with VIP treatement (Default value = [])
        :param VIP_color: str, color to use for VIP data (Default value = None)
        :param include_all_inj_regions: bool (Default value = False)
        :param others_color: str, color for not VIP data (Default value = "white")
    """
    # check coloring mode used and prepare a list COLORS to use for coloring stuff
    if color_by == "manual":
        # check color argument
        if color is None:
            color = brainrender.TRACT_DEFAULT_COLOR
            COLORS = [color for i in range(len(tractography))]
        elif isinstance(color, list):
            if not len(color) == len(tractography):
                raise ValueError(
                    "If a list of colors is passed, it must have the same number of items as the number of tractography traces"
                )
            else:
                for col in color:
                    if not check_colors(col):
                        raise ValueError(
                            "Color variable passed to tractography is invalid: {}".format(
                                col
                            )
                        )

                COLORS = color
        else:
            if not check_colors(color):
                raise ValueError(
                    "Color variable passed to tractography is invalid: {}".format(
                        color
                    )
                )
            else:
                COLORS = [color for i in range(len(tractography))]

    elif color_by == "target_region":
        if VIP_color is not None:
            if not check_colors(VIP_color) or not check_colors(others_color):
                raise ValueError("Invalid VIP or other color passed")
            try:
                if include_all_inj_regions:
                    COLORS = [
                        VIP_color
                        if is_any_item_in_list(
                            [
                                x["abbreviation"]
                                for x in t["injection-structures"]
                            ],
                            VIP_regions,
                        )
                        else others_color
                        for t in tractography
                    ]
                else:
                    COLORS = [
                        VIP_color
                        if t["structure-abbrev"] in VIP_regions
                        else others_color
                        for t in tractography
                    ]
            except:
                raise ValueError(
                    "Something went wrong while getting colors for tractography"
                )
        else:
            COLORS = [None for t in tractography]  # will be filled up later

    else:
        raise ValueError(
            "Unrecognised 'color_by' argument {}".format(color_by)
        )

    return COLORS


def experiments_source_search(mca, SOI, *args, source=True, **kwargs):
    """
        Returns data about experiments whose injection was in the SOI, structure of interest
        :param SOI: str, structure of interest. Acronym of structure to use as seed for teh search
        :param *args: 
        :param source:  (Default value = True)
        :param **kwargs: 
        """
    """
            list of possible kwargs
                injection_structures : list of integers or strings
                    Integer Structure.id or String Structure.acronym.
                target_domain : list of integers or strings, optional
                    Integer Structure.id or String Structure.acronym.
                injection_hemisphere : string, optional
                    'right' or 'left', Defaults to both hemispheres.
                target_hemisphere : string, optional
                    'right' or 'left', Defaults to both hemispheres.
                transgenic_lines : list of integers or strings, optional
                    Integer TransgenicLine.id or String TransgenicLine.name. Specify ID 0 to exclude all TransgenicLines.
                injection_domain : list of integers or strings, optional
                    Integer Structure.id or String Structure.acronym.
                primary_structure_only : boolean, optional
                product_ids : list of integers, optional
                    Integer Product.id
                start_row : integer, optional
                    For paging purposes. Defaults to 0.
                num_rows : integer, optional
                    For paging purposes. Defaults to 2000.
        """
    transgenic_id = kwargs.pop(
        "transgenic_id", 0
    )  # id = 0 means use only wild type
    primary_structure_only = kwargs.pop("primary_structure_only", True)

    if not isinstance(SOI, list):
        SOI = [SOI]

    if source:
        injection_structures = SOI
        target_domain = None
    else:
        injection_structures = None
        target_domain = SOI

    return pd.DataFrame(
        mca.experiment_source_search(
            injection_structures=injection_structures,
            target_domain=target_domain,
            transgenic_lines=transgenic_id,
            primary_structure_only=primary_structure_only,
        )
    )


# ---------------------------------------------------------------------------- #
#                                  STREAMLINES                                 #
# ---------------------------------------------------------------------------- #
def parse_streamline(
    *args,
    filepath=None,
    data=None,
    show_injection_site=True,
    color="ivory",
    alpha=0.8,
    radius=10,
    **kwargs,
):
    """
        Given a path to a .json file with streamline data (or the data themselves), render the streamline as tubes actors.
        Either  filepath or data should be passed

        :param filepath: str, optional. Path to .json file with streamline data (Default value = None)
        :param data: panadas.DataFrame, optional. DataFrame with streamline data. (Default value = None)
        :param color: str color of the streamlines (Default value = 'ivory')
        :param alpha: float transparency of the streamlines (Default value = .8)
        :param radius: int radius of the streamlines actor (Default value = 10)
        :param show_injection_site: bool, if True spheres are used to render the injection volume (Default value = True)
        :param *args: 
        :param **kwargs: 

    """
    if filepath is not None and data is None:
        data = load_json(filepath)
    elif filepath is None and data is not None:
        pass
    else:
        raise ValueError(
            "Need to pass eiteher a filepath or data argument to parse_streamline"
        )

    # create actors for streamlines
    lines = []
    if len(data["lines"]) == 1:
        try:
            lines_data = data["lines"][0]
        except KeyError:
            lines_data = data["lines"]["0"]
    else:
        lines_data = data["lines"]
    for line in lines_data:
        points = [[l["x"], l["y"], l["z"]] for l in line]
        lines.append(
            shapes.Tube(
                points,
                r=radius,
                c=color,
                alpha=alpha,
                res=brainrender.STREAMLINES_RESOLUTION,
            )
        )

    coords = []
    if show_injection_site:
        if len(data["injection_sites"]) == 1:
            try:
                injection_data = data["injection_sites"][0]
            except KeyError:
                injection_data = data["injection_sites"]["0"]
        else:
            injection_data = data["injection_sites"]

        for inj in injection_data:
            coords.append(list(inj.values()))
        spheres = [shapes.Spheres(coords, r=brainrender.INJECTION_VOLUME_SIZE)]
    else:
        spheres = []

    merged = merge(*lines, *spheres)
    merged.color(color)
    merged.alpha(alpha)
    return [merged]


def make_url_given_id(expid):
    """
        Get url of JSON file for an experiment, give it's ID number

        :param expid: int with experiment ID number

    """
    return "https://neuroinformatics.nl/HBP/allen-connectivity-viewer/json/streamlines_{}.json.gz".format(
        expid
    )


def download_streamlines(eids, streamlines_folder=None):  # pragma: no cover
    """
        Given a list of expeirmental IDs, it downloads the streamline data from the https://neuroinformatics.nl cache and saves them as
        json files. 

        :param eids: list of integers with experiments IDs
        :param streamlines_folder: str path to the folder where the JSON files should be saved, if None the default is used (Default value = None)

    """
    streamlines_folder = Path(streamlines_folder)

    if not isinstance(eids, (list, np.ndarray, tuple)):
        eids = [eids]

    filepaths, data = [], []
    for eid in track(eids, total=len(eids), description="downloading"):
        url = make_url_given_id(eid)
        jsonpath = streamlines_folder / f"{eid}.json"
        filepaths.append(str(jsonpath))

        if not jsonpath.exists():
            response = request(url)

            # Write the response content as a temporary compressed file
            temp_path = streamlines_folder / "temp.gz"
            with open(str(temp_path), "wb") as temp:
                temp.write(response.content)

            # Open in pandas and delete temp
            url_data = pd.read_json(
                str(temp_path), lines=True, compression="gzip"
            )
            temp_path.unlink()

            # save json
            url_data.to_json(str(jsonpath))

            # append to lists and return
            data.append(url_data)
        else:
            data.append(pd.read_json(str(jsonpath)))
    return filepaths, data


def extract_ids_from_csv(
    csv_file, download=False, **kwargs
):  # pragma: no cover
    """
        Parse CSV file to extract experiments IDs and link to downloadable streamline data
    
        Given a CSV file with info about experiments downloaded from: http://connectivity.brain-map.org
        extract experiments ID and get links to download (compressed) streamline data from https://neuroinformatics.nl.
        Also return the experiments IDs to download data from: https://neuroinformatics.nl/HBP/allen-connectivity-viewer/streamline-downloader.html
        

        :param csv_file: str with path to csv file
        :param download: if True the data are downloaded automatically (Default value = False)
        :param **kwargs: 

    """
    data = pd.read_csv(csv_file)

    if not download:
        print("To download compressed data, click on the following URLs:")
        for eid in data.id.values:
            url = make_url_given_id(eid)
            print(url)

        print("\n")
        string = ""
        for x in data.id.values:
            string += "{},".format(x)

        print(
            "To download JSON directly, go to: https://neuroinformatics.nl/HBP/allen-connectivity-viewer/streamline-downloader.html"
        )
        print(
            "and  copy and paste the following experiments ID in the 'Enter the Allen Connectivity Experiment number:' field."
        )
        print(
            "You can copy and paste each individually or a list of IDs separated by a comma"
        )
        print("IDs: {}".format(string[:-1]))
        print("\n")

        return data.id.values
    else:
        return download_streamlines(data.id.values, **kwargs)


# ---------------------------------------------------------------------------- #
#                                  SHARPTRACK                                  #
# ---------------------------------------------------------------------------- #


def parse_sharptrack(
    atlas,
    probe_points_file,
    name,
    color_by_region=True,
    color="salmon",
    radius=30,
    probe_color="blackboard",
    probe_radius=15,
    probe_alpha=1,
):
    """
            Visualises the position of an implanted probe in the brain. 
            Uses the location of points along the probe extracted with SharpTrack
            [https://github.com/cortex-lab/allenCCF].
            It renders the position of points along the probe and a line fit through them.
            Code contributed by @tbslv on github. 
        """

    # Points params
    params = dict(color_by_region=True, color="salmon", radius=30, res=12,)

    # Get the position of probe points and render
    probe_points_df = get_probe_points_from_sharptrack(probe_points_file)

    # Fit a line through the points [adapted from SharpTrack by @tbslv]
    r0 = np.mean(probe_points_df.values, axis=0)
    coords = probe_points_df.values - r0
    U, S, V = np.linalg.svd(coords)
    direction = V.T[:, 0]

    # Find intersection with brain surface
    root_mesh = atlas._get_structure_mesh("root")
    p0 = direction * np.array([-1]) + r0
    p1 = (
        direction * np.array([-15000]) + r0
    )  # end point way outside of brain, on probe trajectory though
    pts = root_mesh.intersectWithLine(p0, p1)

    # Define top/bottom coordinates to render as a cylinder
    top_coord = pts[0]
    length = np.sqrt(np.sum((probe_points_df.values[-1] - top_coord) ** 2))
    bottom_coord = top_coord + direction * length

    # Render probe as a cylinder
    probe = Cylinder(
        [top_coord, bottom_coord],
        r=probe_radius,
        alpha=probe_alpha,
        c=probe_color,
    )
    probe.name = name

    return probe_points_df, params, probe, color
